"""Test functionalities of reaction and gene deletions."""

import math
import os
from typing import Callable, List

import numpy as np
import pandas as pd
import pytest

from cobra import Model, Solution
from cobra.flux_analysis.deletion import (
    double_gene_deletion,
    double_reaction_deletion,
    single_gene_deletion,
    single_reaction_deletion,
)
from cobra.flux_analysis.room import add_room


# Single gene deletion FBA
def test_single_gene_deletion_fba_benchmark(
    model: Model, benchmark: Callable, all_solvers: List[str]
) -> None:
    """Benchmark single gene deletion using FBA."""
    model.solver = all_solvers
    benchmark(single_gene_deletion, model, model.genes[1::10])


def test_single_gene_deletion_fba(model: Model, all_solvers: List[str]) -> None:
    """Test single gene deletion using FBA."""
    # expected knockouts for textbook model
    model.solver = all_solvers
    growth_dict = {
        "b0008": 0.87,
        "b0114": 0.80,
        "b0116": 0.78,
        "b2276": 0.21,
        "b1779": 0.00,
    }
    result = single_gene_deletion(
        model=model, gene_list=list(growth_dict), method="fba", processes=1
    )
    for gene, value in growth_dict.items():
        assert np.isclose(result.knockout[gene].growth, value, atol=1e-02)


# Singe gene deletion MOMA
def test_single_gene_deletion_moma_benchmark(
    model: Model, benchmark: Callable, qp_solvers: List[str]
) -> None:
    """Benchmark single gene deletion using MOMA."""
    model.solver = qp_solvers
    genes = ["b0008", "b0114", "b2276", "b1779"]
    benchmark(
        single_gene_deletion,
        model=model,
        gene_list=genes,
        method="moma",
        processes=1,
    )


def test_single_gene_deletion_moma(model: Model, qp_solvers: List[str]) -> None:
    """Test single gene deletion using MOMA."""
    model.solver = qp_solvers
    # expected knockouts for textbook model
    growth_dict = {
        "b0008": 0.87,
        "b0114": 0.71,
        "b0116": 0.56,
        "b2276": 0.11,
        "b1779": 0.00,
    }

    result = single_gene_deletion(
        model=model, gene_list=list(growth_dict), method="moma", processes=1
    )
    for gene, value in growth_dict.items():
        assert np.isclose(result.knockout[gene].growth, value, atol=1e-02)


def test_single_gene_deletion_moma_reference(
    model: Model, qp_solvers: List[str]
) -> None:
    """Test single gene deletion using MOMA (reference solution)."""
    model.solver = qp_solvers
    # expected knockouts for textbook model
    growth_dict = {
        "b0008": 0.87,
        "b0114": 0.71,
        "b0116": 0.56,
        "b2276": 0.11,
        "b1779": 0.00,
    }

    sol = model.optimize()
    result = single_gene_deletion(
        model=model,
        gene_list=list(growth_dict),
        method="moma",
        solution=sol,
        processes=1,
    )
    for gene, value in growth_dict.items():
        assert np.isclose(result.knockout[gene].growth, value, atol=1e-02)


# Single gene deletion linear MOMA
def test_single_gene_deletion_linear_moma_benchmark(
    model: Model, benchmark: Callable, all_solvers: List[str]
) -> None:
    """Benchmark single gene deletion using linear MOMA."""
    model.solver = all_solvers
    genes = ["b0008", "b0114", "b2276", "b1779"]
    benchmark(
        single_gene_deletion,
        model=model,
        gene_list=genes,
        method="linear moma",
        processes=1,
    )


def test_single_gene_deletion_linear_moma(model: Model, all_solvers: List[str]) -> None:
    """Test single gene deletion using linear MOMA (reference solution)."""
    model.solver = all_solvers
    # expected knockouts for textbook model
    growth_dict = {
        "b0008": 0.87,
        "b0114": 0.76,
        "b0116": 0.65,
        "b2276": 0.08,
        "b1779": 0.00,
    }

    sol = model.optimize()
    result = single_gene_deletion(
        model=model,
        gene_list=list(growth_dict),
        method="linear moma",
        solution=sol,
        processes=1,
    )
    for gene, value in growth_dict.items():
        assert np.isclose(result.knockout[gene].growth, value, atol=1e-02)


# Single gene deletion ROOM
def test_single_gene_deletion_room_benchmark(
    model: Model, benchmark: Callable, all_solvers: List[str]
) -> None:
    """Benchmark single gene deletion using ROOM."""
    if all_solvers == "glpk":
        pytest.skip("GLPK is too slow to run ROOM.")
    model.solver = all_solvers
    genes = ["b0008", "b0114", "b2276", "b1779"]
    benchmark(
        single_gene_deletion,
        model=model,
        gene_list=genes,
        method="room",
        processes=1,
    )


# Single gene deletion linear ROOM
def test_single_gene_deletion_linear_room_benchmark(
    model: Model, benchmark: Callable, all_solvers: List[str]
) -> None:
    """Benchmark single gene deletion using linear ROOM."""
    model.solver = all_solvers
    genes = ["b0008", "b0114", "b2276", "b1779"]
    benchmark(
        single_gene_deletion,
        model=model,
        gene_list=genes,
        method="linear room",
        processes=1,
    )


# Single reaction deletion
def test_single_reaction_deletion_benchmark(
    model: Model, benchmark: Callable, all_solvers: List[str]
) -> None:
    """Benchmark single reaction deletion."""
    model.solver = all_solvers
    benchmark(
        single_reaction_deletion, model=model, genes=model.genes[1::10], processes=1
    )


def test_single_reaction_deletion(model: Model, all_solvers) -> None:
    """Test single reaction deletion."""
    model.solver = all_solvers
    expected_results = {
        "FBA": 0.70404,
        "FBP": 0.87392,
        "CS": 0,
        "FUM": 0.81430,
        "GAPD": 0,
        "GLUDy": 0.85139,
    }
    result = single_reaction_deletion(
        model=model, reaction_list=list(expected_results), processes=1
    )

    for reaction, value in expected_results.items():
        assert np.isclose(result.knockout[reaction].growth, value, atol=1e-05)


# Single reaction deletion ROOM
def test_single_reaction_deletion_room(
    room_model: Model, room_solution: Solution, all_solvers: List[str]
) -> None:
    """Test single reaction deletion using ROOM."""
    room_model.solver = all_solvers
    expected = pd.Series(
        {
            "v1": 10.0,
            "v2": 5.0,
            "v3": 0.0,
            "v4": 5.0,
            "v5": 5.0,
            "v6": 0.0,
            "b1": 10.0,
            "b2": 5.0,
            "b3": 5.0,
        },
        index=["v1", "v2", "v3", "v4", "v5", "v6", "b1", "b2", "b3"],
    )
    with room_model:
        room_model.reactions.v6.knock_out()
        add_room(room_model, solution=room_solution, delta=0.0, epsilon=0.0)
        room_sol = room_model.optimize()

    assert np.allclose(room_sol.fluxes, expected)


# Single reaction deletion linear ROOM
def test_single_reaction_deletion_linear_room(
    room_model: Model, room_solution: Solution, all_solvers: List[str]
) -> None:
    """Test single reaction deletion using linear ROOM."""
    room_model.solver = all_solvers
    expected = pd.Series(
        {
            "v1": 10.0,
            "v2": 5.0,
            "v3": 0.0,
            "v4": 5.0,
            "v5": 5.0,
            "v6": 0.0,
            "b1": 10.0,
            "b2": 5.0,
            "b3": 5.0,
        },
        index=["v1", "v2", "v3", "v4", "v5", "v6", "b1", "b2", "b3"],
    )
    with room_model:
        room_model.reactions.v6.knock_out()
        add_room(
            room_model,
            solution=room_solution,
            delta=0.0,
            epsilon=0.0,
            linear=True,
        )
        linear_room_sol = room_model.optimize()

    assert np.allclose(linear_room_sol.fluxes, expected)


# Double gene deletion
def test_double_gene_deletion_benchmark(
    large_model: Model, benchmark: Callable
) -> None:
    """Benchmark double gene deletion."""
    genes = [
        "b0726",
        "b4025",
        "b0724",
        "b0720",
        "b2935",
        "b2935",
        "b1276",
        "b1241",
    ]
    benchmark(double_gene_deletion, large_model, gene_list1=genes, processes=1)


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_double_gene_deletion(model: Model) -> None:
    """Test double gene deletion."""
    genes = [
        "b0726",
        "b4025",
        "b0724",
        "b0720",
        "b2935",
        "b2935",
        "b1276",
        "b1241",
    ]
    growth_dict = {
        "b0720": {
            "b0720": 0.0,
            "b0724": 0.0,
            "b0726": 0.0,
            "b1241": 0.0,
            "b1276": 0.0,
            "b2935": 0.0,
            "b4025": 0.0,
        },
        "b0724": {
            "b0720": 0.0,
            "b0724": 0.814,
            "b0726": 0.814,
            "b1241": 0.814,
            "b1276": 0.814,
            "b2935": 0.814,
            "b4025": 0.739,
        },
        "b0726": {
            "b0720": 0.0,
            "b0724": 0.814,
            "b0726": 0.858,
            "b1241": 0.858,
            "b1276": 0.858,
            "b2935": 0.858,
            "b4025": 0.857,
        },
        "b1241": {
            "b0720": 0.0,
            "b0724": 0.814,
            "b0726": 0.858,
            "b1241": 0.874,
            "b1276": 0.874,
            "b2935": 0.874,
            "b4025": 0.863,
        },
        "b1276": {
            "b0720": 0.0,
            "b0724": 0.814,
            "b0726": 0.858,
            "b1241": 0.874,
            "b1276": 0.874,
            "b2935": 0.874,
            "b4025": 0.863,
        },
        "b2935": {
            "b0720": 0.0,
            "b0724": 0.814,
            "b0726": 0.858,
            "b1241": 0.874,
            "b1276": 0.874,
            "b2935": 0.874,
            "b4025": 0.863,
        },
        "b4025": {
            "b0720": 0.0,
            "b0724": 0.739,
            "b0726": 0.857,
            "b1241": 0.863,
            "b1276": 0.863,
            "b2935": 0.863,
            "b4025": 0.863,
        },
    }
    solution = double_gene_deletion(model, gene_list1=genes, processes=3)
    solution_one_process = double_gene_deletion(model, gene_list1=genes, processes=1)

    for rxn_a, sub in growth_dict.items():
        for rxn_b, growth in sub.items():
            sol = solution.knockout[{rxn_a, rxn_b}]
            sol_one = solution_one_process.knockout[{rxn_a, rxn_b}]
            assert np.isclose(sol.growth, growth, atol=1e-3)
            assert np.isclose(sol_one.growth, growth, atol=1e-3)


def test_double_gene_knockout_bug(large_model: Model) -> None:
    """Test that the bug reported in #1102 is fixed."""
    genes = ["b0118", "b1276"]
    expected = (
        pd.DataFrame(
            data={
                "ids": [
                    {"b0118"},
                    {"b1276"},
                    {"b1276", "b0118"},
                ],
                "growth": [0.98, 0.98, 0.0],
                "status": ["optimal"] * 3,
            }
        )
        .sort_values("ids")
        .reset_index()
    )
    result = (
        double_gene_deletion(large_model, genes, processes=1)
        .sort_values("ids")
        .reset_index()
    )
    assert result["growth"].values == pytest.approx(expected["growth"].values, abs=0.01)
    assert (result["status"] == expected["status"]).all()


# Double reaction deletion
def test_double_reaction_deletion_benchmark(
    large_model: Model, benchmark: Callable
) -> None:
    """Benchmark double reaction deletion."""
    reactions = large_model.reactions[100:105]
    benchmark(double_reaction_deletion, large_model, reaction_list1=reactions)


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_double_reaction_deletion(model: Model) -> None:
    """Test double reaction deletion."""
    reactions = ["FBA", "ATPS4r", "ENO", "FRUpts2"]
    growth_dict = {
        "FBA": {"ATPS4r": 0.135, "ENO": float("nan"), "FRUpts2": 0.704},
        "ATPS4r": {"ENO": float("nan"), "FRUpts2": 0.374},
        "ENO": {"FRUpts2": 0.0},
    }

    solution = double_reaction_deletion(model, reaction_list1=reactions, processes=3)
    solution_one_process = double_reaction_deletion(
        model, reaction_list1=reactions, processes=1
    )
    for rxn_a, sub in growth_dict.items():
        for rxn_b, growth in sub.items():
            sol = solution.knockout[{rxn_a, rxn_b}]
            sol_one = solution_one_process.knockout[{rxn_a, rxn_b}]
            if math.isnan(growth):
                assert math.isnan(sol.growth)
                assert math.isnan(sol_one.growth)
            else:
                assert np.isclose(sol.growth, growth, atol=1e-3)
                assert np.isclose(sol_one.growth, growth, atol=1e-3)


def test_deletion_accessor(small_model: Model) -> None:
    """Test the DataFrame accessor."""
    single = single_reaction_deletion(small_model, small_model.reactions[0:10])
    double = double_reaction_deletion(small_model, small_model.reactions[0:10])
    rxn1 = small_model.reactions[0]
    rxn2 = small_model.reactions[1]

    with pytest.raises(ValueError):
        single.knockout[1]

    with pytest.raises(ValueError):
        single.knockout[{"a": 1}]

    assert single.knockout[rxn1].ids.iloc[0] == {rxn1.id}
    assert double.knockout[{rxn1, rxn2}].ids.iloc[0] == {rxn1.id, rxn2.id}
    assert all(single.knockout[rxn1.id] == single.knockout[rxn1])
    assert all(double.knockout[{rxn1.id, rxn2.id}] == double.knockout[{rxn1, rxn2}])
    assert single.knockout[rxn1, rxn2].shape == (2, 3)
    assert double.knockout[rxn1, rxn2].shape == (2, 3)
    assert double.knockout[{rxn1, rxn2}].shape == (1, 3)
    assert double.knockout[{rxn1}, {rxn2}].shape == (2, 3)



"""Test functions of solution.py ."""

from typing import TYPE_CHECKING, Tuple

from cobra.core import Solution


if TYPE_CHECKING:
    from cobra import Model


def test_solution_contains_only_reaction_specific_values(
    solved_model: Tuple[Solution, "Model"],
) -> None:
    """Test solution contains specific reaction values."""
    solution, model = solved_model
    reaction_ids = set([reaction.id for reaction in model.reactions])
    assert set(solution.fluxes.index) == reaction_ids
    # assert set(solution.reduced_costs.index) == reaction_ids



"""Test functions of context.py."""

from typing import TYPE_CHECKING

from cobra.util import HistoryManager, get_context, resettable


if TYPE_CHECKING:
    from cobra import Model


def test_history_manager() -> None:
    """Test initialization and resetting of HistoryManager."""
    # initialize HistoryManager
    history_manager = HistoryManager()
    # add non-functioning operation
    history_manager(lambda: 1)
    # check size of the stack
    assert history_manager.size() == 1
    # reset operations
    history_manager.reset()


def test_get_context(model: "Model") -> None:
    """Test if context retrieval is working."""
    with model as model:
        # reverse optimization direcion
        model.objective_direction = "min"
        # retrieve context
        context = get_context(model)
        # check size of the context
        if context:
            assert context.size() == 1

    # there shouldn't be any history
    if context:
        assert context.size() == 0


def test_resettable() -> None:
    """Test if resettable decorator is functional."""

    # decorate a dummy function
    @resettable
    def change_my_name(old_name, new_name):
        """Change old name to new name."""
        if old_name != new_name:
            old_name = new_name

    # call the dummy function
    change_my_name("hmm", "hmmm")



"""Define module level fixtures."""

import json
from os.path import join
from typing import List

import pandas as pd
import pytest

from cobra.core import Model, Reaction, Solution
from cobra.util import solver as sutil


# The scipy interface is currently unstable and may yield errors or infeasible
# solutions.
@pytest.fixture(
    scope="session",
    params=[s for s in ["glpk", "cplex", "gurobi", "hybrid"] if s in sutil.solvers],
)
def all_solvers(request) -> List[str]:
    """Return the avaialble solvers."""
    return request.param


@pytest.fixture(
    scope="session",
    params=[s for s in ["cplex", "gurobi", "hybrid"] if s in sutil.solvers],
)
def qp_solvers(request) -> List[str]:
    """Return the available QP solvers."""
    return request.param


@pytest.fixture(scope="module")
def fva_results(data_directory) -> pd.DataFrame:
    """Load and return saved FVA results for textbook model."""
    with open(join(data_directory, "textbook_fva.json"), "r") as infile:
        df = pd.DataFrame(json.load(infile))
    df.sort_index(inplace=True)
    return df[["minimum", "maximum"]]


@pytest.fixture(scope="module")
def pfba_fva_results(data_directory) -> pd.DataFrame:
    """Load and return saved pFBA FVA results for textbook model."""
    with open(join(data_directory, "textbook_pfba_fva.json"), "r") as infile:
        df = pd.DataFrame(json.load(infile))
    df.sort_index(inplace=True)
    return df[["minimum", "maximum"]]


@pytest.fixture(scope="module")
def room_model() -> Model:
    """Generate ROOM model as described in [1]_ .

    References
    ----------
    .. [1] Tomer Shlomi, Omer Berkman and Eytan Ruppin, "Regulatory on/off
     minimization of metabolic flux changes after genetic perturbations",
     PNAS 2005 102 (21) 7695-7700; doi:10.1073/pnas.0406346102

    """
    test_model = Model("papin_2003")

    v_1 = Reaction("v1")
    v_2 = Reaction("v2")
    v_3 = Reaction("v3")
    v_4 = Reaction("v4")
    v_5 = Reaction("v5")
    v_6 = Reaction("v6", upper_bound=0.0)
    b_1 = Reaction("b1", upper_bound=10.0, lower_bound=0.0)
    b_2 = Reaction("b2")
    b_3 = Reaction("b3")

    test_model.add_reactions([v_1, v_2, v_3, v_4, v_5, v_6, b_1, b_2, b_3])

    v_1.reaction = "A -> B"
    v_2.reaction = "2 B -> C + byp"
    v_3.reaction = "2 B + cof -> D"
    v_4.reaction = "D -> E + cof"
    v_5.reaction = "C + cof -> D"
    v_6.reaction = "C -> E"
    b_1.reaction = "-> A"
    b_2.reaction = "E ->"
    b_3.reaction = "byp ->"

    test_model.objective = "b2"

    return test_model


@pytest.fixture(scope="module")
def room_solution() -> Solution:
    """Generate ROOM solution as described in [1]_ .

    References
    ----------
    .. [1] Tomer Shlomi, Omer Berkman and Eytan Ruppin, "Regulatory on/off
     minimization of metabolic flux changes after genetic perturbations",
     PNAS 2005 102 (21) 7695-7700; doi:10.1073/pnas.0406346102

    """
    fluxes = pd.Series(
        {
            "b1": 10.0,
            "b2": 5.0,
            "b3": 5.0,
            "v1": 10.0,
            "v2": 5.0,
            "v3": 0.0,
            "v4": 0.0,
            "v5": 0.0,
            "v6": 5.0,
        }
    )
    reduced_costs = pd.Series(
        {
            "b1": 0.0,
            "b2": 0.0,
            "b3": 0.0,
            "v1": 0.0,
            "v2": 0.0,
            "v3": 0.0,
            "v4": 0.0,
            "v5": 0.0,
            "v6": 0.0,
        }
    )
    shadow_prices = pd.Series(
        {
            "b1": 0.0,
            "b2": 0.0,
            "b3": 0.0,
            "v1": 0.0,
            "v2": 0.0,
            "v3": 0.0,
            "v4": 0.0,
            "v5": 0.0,
            "v6": 0.0,
        }
    )
    sol = Solution(
        objective_value=5.000,
        status="optimal",
        fluxes=fluxes,
        reduced_costs=reduced_costs,
        shadow_prices=shadow_prices,
    )
    return sol



"""Test functions of util.py ."""

import pytest

from cobra.util import AutoVivification, format_long_string, show_versions


@pytest.mark.parametrize(
    "input_string, expected_string",
    [
        (
            (
                "This is a really long string, but is it long enough. "
                "I hope it is long enough so that format_long_string() function works."
            ),
            "This is a really long string, but is it long...",
        ),
        ("This is short string.", "This is short string."),
    ],
)
def test_format_long_string(input_string: str, expected_string: str) -> None:
    """Test functionality of format long string."""
    assert expected_string == format_long_string(input_string)


def test_autovivification() -> None:
    """Test proper functionality of autovivification."""
    test_data = AutoVivification()
    test_data["a"]["b"] = 1
    test_data["c"]["d"] = 2
    assert test_data["a"] == {"b": 1}
    assert test_data["c"] == {"d": 2}
    assert test_data["a"]["b"] == 1
    assert test_data["c"]["d"] == 2


def test_show_versions(capsys) -> None:
    """Test output of dependency information."""
    show_versions()
    captured = capsys.readouterr()
    lines = captured.out.split("\n")
    assert lines[1].startswith("Package Information")
    assert lines[2].startswith("------------------")
    assert lines[3].startswith("cobra")

    assert lines[5].startswith("Dependency Information")
    assert lines[6].startswith("------------------")



"""Provides Fast-SNP algorithm implementation for finding nullspace basis of matrix."""

from typing import List, Tuple

import numpy as np
import optlang
from optlang.interface import OPTIMAL, Model, Variable
from optlang.symbolics import Zero


def _solve_snv(
    weights: np.ndarray, model: Model, v_list: List[Variable], positive: bool
) -> np.ndarray:
    """Optimize Fast-SNP step for given weights and direction."""

    dir = 1 if positive else -1

    model.constraints["nonzero_constraint"].set_linear_coefficients(
        {variable: weight * dir for variable, weight in zip(v_list, weights)}
    )

    model.optimize()
    if model.status != OPTIMAL:
        return None

    result = np.array([variable.primal for variable in v_list])
    result /= np.linalg.norm(result)

    return result


def _create_fast_snp_problem(
    solver: "optlang.interface",
    S: np.ndarray,
    directions: np.ndarray,
    v_bound: float,
    zero_cutoff: float,
    bias: float,
) -> Tuple[Model, List[Variable]]:
    """Create an optimization problem for Fast-SNP algorithm."""

    n = S.shape[1]

    model = solver.Model()
    modulo_list = []
    v_list = []

    for i in range(n):
        v = solver.Variable(
            name=f"v_{i}",
            lb=v_bound * min(directions[i, 0], 0),
            ub=v_bound * max(directions[i, 1], 0),
            problem=model,
        )
        model.add([v])
        v_list.append(v)

        if directions[i, 0] == 0:
            modulo_list.append((v, 1))
        elif directions[i, 1] == 0:
            modulo_list.append((v, -1))
        else:
            x = solver.Variable(name=f"x_{i}", lb=0, problem=model)
            model.add([x])
            modulo_list.append((x, 1))

            constraint1 = solver.Constraint(
                Zero,
                lb=0.0,
                name=f"modulo_constraint1_{i}",
            )
            constraint2 = solver.Constraint(
                Zero,
                lb=0.0,
                name=f"modulo_constraint2_{i}",
            )
            model.add([constraint1, constraint2], sloppy=True)

            model.constraints[f"modulo_constraint1_{i}"].set_linear_coefficients(
                {x: 1.0, v: -1.0}
            )
            model.constraints[f"modulo_constraint2_{i}"].set_linear_coefficients(
                {x: 1.0, v: 1.0}
            )

    for idx, row in enumerate(S):
        nnz_list = np.flatnonzero(np.abs(row) > zero_cutoff)
        if len(nnz_list) == 0:
            continue

        constraint = solver.Constraint(Zero, lb=0.0, ub=0.0, name=f"row_{idx}")
        model.add([constraint], sloppy=True)
        model.constraints[f"row_{idx}"].set_linear_coefficients(
            {v_list[i]: row[i] for i in nnz_list}
        )

    model.add(
        [
            solver.Constraint(
                Zero,
                lb=bias,
                name="nonzero_constraint",
            )
        ],
        sloppy=True,
    )

    model.objective = solver.Objective(Zero)
    model.objective.set_linear_coefficients({x: coef for (x, coef) in modulo_list})
    model.objective.direction = "min"

    return model, v_list


def _project(N: np.ndarray, w: np.ndarray) -> np.ndarray:
    """Project vector w to the complement of the row space of N."""
    return w - (N.T @ (N @ w))


def _get_condition_vector(N: np.ndarray) -> np.ndarray:
    """Get a random Fast-SNP nonzero condition vector."""
    return _project(N, np.random.normal(size=N.shape[1]))


def nullspace_fast_snp(
    solver: "optlang.interface",
    S: np.ndarray,
    directions: np.ndarray,
    v_bound: float = 1e4,
    zero_cutoff: float = 1e-6,
    bias: float = 1,
    required_stop_checks_num: int = 3,
) -> np.ndarray:
    """Compute an approximate basis for the nullspace of S with coordinate directions.

    The algorithm used by this function is described in [1]_.

    Parameters
    ----------
    solver : "optlang.interface"
        The solver interface to use for the optimization problem.
        You can use `model.problem` to get the solver interface.
    S : numpy.ndarray
        The matrix for which the nullspace is computed.
        `S` should be a 2-D array.
    directions : numpy.ndarray
        A 2-D array with shape (k, 2) where `k` is the number of columns in `S`.
        This array specifies the directions of coordinates.
        Each row should be:
            - [0, 0] for coordinates that can be only zero
            - [0, 1] for coordinates that can be only positive
            - [-1, 0] for coordinates that can be only negative
            - [-1, 1] for coordinates that can be both positive and negative
    v_bound : float, optional
        The bound for the variables in the optimization problem (default 1e4).
    zero_cutoff : float, optional
        The cutoff value to consider a coordinate value as zero (default 1e-6).
    bias : float, optional
        The bias for the non-zero constraint in the optimization problem
        (default 1).
    required_stop_checks_num : int, optional
        The number of random checks to pass to prove that basis
        could not be expanded (default 3).

    Returns
    -------
    numpy.ndarray
        If `S` is an array with shape (m, k), then an array
        with shape (k, n) will be returned, where `n` is the dimension of the
        nullspace of `S` with `directions`. Each column of this array is a basis
        vector for the nullspace; each element in numpy.dot(S, column) will be
        approximately zero. Each coordinate in the column will have an allowed
        sign according to the `directions` parameter.

    References
    ----------
    .. [1] Fast-SNP: a fast matrix pre-processing algorithm for efficient
       loopless flux optimization of metabolic models. Saa PA, Nielsen LK.
       Bioinformatics. 2016 Dec;32(24):3807–3814. doi: 10.1093/bioinformatics/btw555.
    """

    if len(S.shape) != 2 or S.shape[0] == 0 or S.shape[1] == 0:
        raise ValueError("Input matrix S must be a 2D array with non-zero dimensions.")

    problem, v_list = _create_fast_snp_problem(
        solver, S, directions, v_bound, zero_cutoff, bias
    )

    n = S.shape[1]
    N = np.zeros((0, n))
    U = np.zeros((0, n))
    stop_checks_num = 0

    while N.shape[0] < n and stop_checks_num < required_stop_checks_num:
        weights = _get_condition_vector(U)

        v1 = _solve_snv(weights, problem, v_list, True)
        v2 = _solve_snv(weights, problem, v_list, False)

        if v1 is None and v2 is None:
            stop_checks_num += 1
            continue

        stop_checks_num = 0
        v = v1
        if v1 is None:
            v = v2

        if v1 is not None and v2 is not None:
            nnz_v1 = np.sum(np.abs(v1) > zero_cutoff)
            nnz_v2 = np.sum(np.abs(v2) > zero_cutoff)
            if nnz_v1 > nnz_v2:
                v = v2

        N = np.vstack([N, v])

        v = _project(U, v)
        v /= np.linalg.norm(v)
        U = np.vstack([U, v])

    return N.T



"""Test functionalities of FASTCC."""

from typing import Callable, List

import pytest

from cobra import Model, Reaction
from cobra.flux_analysis import fastcc, find_blocked_reactions


@pytest.fixture(scope="module")
def figure1_model() -> Model:
    """Generate a toy model as described in [1]_ figure 1.

    References
    ----------
    .. [1] Vlassis N, Pacheco MP, Sauter T (2014)
           Fast Reconstruction of Compact Context-Specific Metabolic Network
           Models.
           PLoS Comput Biol 10(1): e1003424. doi:10.1371/journal.pcbi.1003424

    """
    test_model = Model("figure 1")
    v1 = Reaction("v1")
    v2 = Reaction("v2")
    v3 = Reaction("v3")
    v4 = Reaction("v4")
    v5 = Reaction("v5")
    v6 = Reaction("v6")

    test_model.add_reactions([v1, v2, v3, v4, v5, v6])

    v1.reaction = "-> 2 A"
    v2.reaction = "A <-> B"
    v3.reaction = "A -> D"
    v4.reaction = "A -> C"
    v5.reaction = "C -> D"
    v6.reaction = "D ->"

    v1.bounds = (0.0, 3.0)
    v2.bounds = (-3.0, 3.0)
    v3.bounds = (0.0, 3.0)
    v4.bounds = (0.0, 3.0)
    v5.bounds = (0.0, 3.0)
    v6.bounds = (0.0, 3.0)

    test_model.objective = v6
    return test_model


@pytest.fixture(scope="module")
def opposing_model() -> Model:
    """Generate a toy model with opposing reversible reactions.

    This toy model ensures that two opposing reversible reactions do not
    appear as blocked.

    """
    test_model = Model("opposing")
    v1 = Reaction("v1")
    v2 = Reaction("v2")
    v3 = Reaction("v3")
    v4 = Reaction("v4")

    test_model.add_reactions([v1, v2, v3, v4])

    v1.reaction = "-> 2 A"
    v2.reaction = "A -> C"  # Later made reversible via bounds.
    v3.reaction = "D -> C"  # Later made reversible via bounds.
    v4.reaction = "D ->"

    v1.bounds = 0.0, 3.0
    v2.bounds = -3.0, 3.0
    v3.bounds = -3.0, 3.0
    v4.bounds = 0.0, 3.0

    test_model.objective = v4
    return test_model


def test_fastcc_benchmark(
    model: Model, benchmark: Callable, all_solvers: List[str]
) -> None:
    """Benchmark fastcc."""
    model.solver = all_solvers
    benchmark(fastcc, model)


def test_figure1(figure1_model: Model, all_solvers: List[str]) -> None:
    """Test FASTCC."""
    figure1_model.solver = all_solvers
    consistent_model = fastcc(figure1_model)
    expected_reactions = {"v1", "v3", "v4", "v5", "v6"}
    assert expected_reactions == {rxn.id for rxn in consistent_model.reactions}


def test_opposing(opposing_model: Model, all_solvers: List[str]) -> None:
    """Test FASTCC."""
    opposing_model.solver = all_solvers
    consistent_model = fastcc(opposing_model)
    expected_reactions = {"v1", "v2", "v3", "v4"}
    assert expected_reactions == {rxn.id for rxn in consistent_model.reactions}


def test_fastcc_against_nonblocked_rxns(model: Model, all_solvers: List[str]) -> None:
    """Test non-blocked reactions obtained by FASTCC."""
    model.solver = all_solvers
    model.tolerance = 1e-6
    fastcc_consistent_model = fastcc(model, 1e-3, 1e-6)
    blocked = find_blocked_reactions(model)
    fastcc_ids = set(rxn.id for rxn in fastcc_consistent_model.reactions)
    assert len(model.reactions) - len(blocked) == len(fastcc_consistent_model.reactions)
    assert fastcc_ids & set(blocked) == set()


def test_fastcc_against_nonblocked_rxns_large(large_model: Model) -> None:
    """Test non-blocked reactions obtained by FASTCC."""
    model = large_model
    model.tolerance = 1e-7
    fastcc_consistent_model = fastcc(model, 1e-3, 1e-7)
    blocked = find_blocked_reactions(model, zero_cutoff=1e-7)
    fastcc_ids = set(rxn.id for rxn in fastcc_consistent_model.reactions)
    assert len(model.reactions) - len(blocked) == len(fastcc_consistent_model.reactions)
    assert fastcc_ids & set(blocked) == set()



"""Test the behaviour of the ProcessPool class."""

import os
from typing import Iterable, Tuple

import pytest
from pytest_mock import MockerFixture

from cobra.util import ProcessPool


def dummy_initializer(*args: Iterable) -> Tuple:
    """Implement a 'do nothing' function that accepts initialization arguments."""
    return args


def square(num: int) -> int:
    """Return the square of an integer."""
    return num * num


def summation(*args: Iterable[int]) -> int:
    """Return the sum of all integer arguments."""
    return sum(args)


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
@pytest.mark.parametrize(
    "attributes",
    [
        {},
        {"processes": 2},
        {"initializer": dummy_initializer},
        {"initializer": dummy_initializer, "initargs": (1, "2", [3], {"a": 4})},
        {"maxtasksperchild": 1},
    ],
)
def test_init(attributes: dict) -> None:
    """Test that a process pool can be initialized with each of its arguments."""
    ProcessPool(**attributes)


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_close(mocker: MockerFixture) -> None:
    """Test that the composed pool is closed as well."""
    pool = ProcessPool(processes=3)
    mock = mocker.patch.object(pool, "_pool", autospec=True)
    pool.close()
    mock.close.assert_called_once()


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_with_context(mocker: MockerFixture) -> None:
    """Test that the composed pool's context is managed as well."""
    pool = ProcessPool(processes=3)
    mock = mocker.patch.object(pool, "_pool", autospec=True)
    with pool:
        pass
    mock.__enter__.assert_called_once()
    mock.__exit__.assert_called_once()


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_apply() -> None:
    """Test that a function can be applied."""
    with ProcessPool(processes=3) as pool:
        assert pool.apply(square, (3,)) == 9


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_apply_async() -> None:
    """Test that a function can be applied asynchronously."""
    with ProcessPool(processes=3) as pool:
        assert pool.apply_async(square, (3,)).get() == 9


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_map() -> None:
    """Test that a function can be mapped over an iterable of values."""
    with ProcessPool(processes=3) as pool:
        assert sum(pool.map(square, [2] * 6)) == 24


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_map_async() -> None:
    """Test that a function can be mapped over an iterable of values asynchronously."""
    with ProcessPool(processes=3) as pool:
        assert sum(pool.map_async(square, [2] * 6).get()) == 24


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_imap() -> None:
    """Test that mapped function results can be iterated."""
    with ProcessPool(processes=3) as pool:
        total = 0
        for result in pool.imap(square, [2] * 6):
            total += result
        assert total == 24


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_imap_unordered() -> None:
    """Test that mapped function results can be iterated in any order."""
    with ProcessPool(processes=3) as pool:
        assert sum(pool.imap_unordered(square, [2] * 6)) == 24


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_starmap() -> None:
    """Test that a function can be starmapped over many iterables."""
    with ProcessPool(processes=3) as pool:
        assert sum(pool.starmap(summation, [range(10), range(10), range(10)])) == 135


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_starmap_async() -> None:
    """Test that a function can be starmapped over many iterables asynchronously."""
    with ProcessPool(processes=3) as pool:
        assert (
            sum(pool.starmap_async(summation, [range(10), range(10), range(10)]).get())
            == 135
        )



"""Test functionalities of I/O in an usage order."""

import logging
from operator import attrgetter
from pathlib import Path
from random import sample
from typing import Iterable, List

import pytest

from cobra import DictList, Model
from cobra import io as cio


logger = logging.getLogger(__name__)


@pytest.fixture(scope="module")
def tmp_path_order(tmp_path_factory: Path) -> Path:
    """Temporary path for I/O order tests."""
    return tmp_path_factory.mktemp("model_order")


@pytest.fixture(scope="module")
def minimized_shuffle(small_model: Model) -> Model:
    """Generate a minimal shuffled model for I/O order tests."""
    model = small_model.copy()
    chosen = sample(list(set(model.reactions) - set(model.exchanges)), 10)
    new = Model("minimized_shuffle")
    new.add_reactions(chosen)
    logger.debug(
        f"'{new.id}' has {new.metabolites} metabolites, {new.reactions} reactions, and "
        f"{new.genes} genes."
    )
    return new


@pytest.fixture(scope="module")
def minimized_sorted(minimized_shuffle: Model) -> Model:
    """Generate a minimal sorted model for I/O order tests."""
    model = minimized_shuffle.copy()
    model.id = "minimized_sorted"
    model.metabolites = DictList(sorted(model.metabolites, key=attrgetter("id")))
    model.genes = DictList(sorted(model.genes, key=attrgetter("id")))
    model.reactions = DictList(sorted(model.reactions, key=attrgetter("id")))
    return model


@pytest.fixture(scope="module")
def minimized_reverse(minimized_shuffle: Model) -> Model:
    """Generate a minimal reversed model for I/O order tests."""
    model = minimized_shuffle.copy()
    model.id = "minimized_reverse"
    model.metabolites = DictList(
        sorted(model.metabolites, key=attrgetter("id"), reverse=True)
    )
    model.genes = DictList(sorted(model.genes, key=attrgetter("id"), reverse=True))
    model.reactions = DictList(
        sorted(model.reactions, key=attrgetter("id"), reverse=True)
    )
    return model


@pytest.fixture(
    scope="module",
    params=["minimized_shuffle", "minimized_reverse", "minimized_sorted"],
)
def template(
    request: pytest.FixtureRequest,
    minimized_shuffle: Model,
    minimized_reverse: Model,
    minimized_sorted: Model,
) -> Model:
    """Return the cobra Model instances found in the current local symbol table."""
    return locals()[request.param]


@pytest.fixture(scope="module", params=["metabolites", "reactions", "genes"])
def attribute(request: pytest.FixtureRequest) -> str:
    """Return the parameter passed."""
    return request.param


def _get_ids(iterable: Iterable) -> List[str]:
    """Get IDs for elements in `iterable`."""
    return [x.id for x in iterable]


@pytest.mark.parametrize(
    "read, write, ext",
    [
        ("read_sbml_model", "write_sbml_model", ".xml"),
        ("load_json_model", "save_json_model", ".json"),
        ("load_yaml_model", "save_yaml_model", ".yml"),
    ],
)
def test_io_order(
    attribute: str,
    read: str,
    write: str,
    ext: str,
    template: Model,
    tmp_path_order: Path,
) -> None:
    """Test loading and saving of models in order.

    Parameters
    ----------
    attribute : str
        The attribute of cobra Model to access.
    read : str
        The function name for loading model, defined as string.
    write : str
        The function name for saving model, defined as string.
    ext : str
        The extension of the file format for loading and saving model.
    template : cobra.Model
        The cobra Model instance to load and save.
    tmp_path_order : pathlib.Path
        The folder path for storing I/O order test files.

    """
    read = getattr(cio, read)
    write = getattr(cio, write)
    file_path = tmp_path_order / f"template{ext}"
    write(template, str(file_path.resolve()))
    model = read(str(file_path.resolve()))
    model_elements = _get_ids(getattr(model, attribute))
    template_elements = _get_ids(getattr(template, attribute))
    assert len(model_elements) == len(template_elements)
    assert set(model_elements) == set(template_elements)
    assert model_elements == template_elements



"""Unit test the model summary."""

import pytest

from cobra.flux_analysis import flux_variability_analysis, pfba
from cobra.summary import ModelSummary


def test_model_summary_interface(model, opt_solver):
    """Test that a summary can be created successfully."""
    model.solver = opt_solver
    ModelSummary(
        model=model,
    )
    ModelSummary(
        model=model,
        solution=pfba(model),
    )
    ModelSummary(
        model=model,
        fva=0.95,
    )
    ModelSummary(
        model=model,
        fva=flux_variability_analysis(
            model, reaction_list=["CYTBD", "NADH16", "SUCDi"]
        ),
    )


def test_model_summary_to_frame(model, opt_solver):
    """Test that the summary's method ``to_frame`` can be called."""
    model.solver = opt_solver
    summary = model.summary()
    summary.to_frame()


@pytest.mark.parametrize(
    "kwargs",
    [
        {},
        {"names": True},
        {"float_format": ".1f"},
        {"threshold": 1e-2},
        {"column_width": 20},
    ],
)
def test_model_summary_to_string(model, opt_solver, kwargs):
    """Test that the summary's method ``to_string`` can be called."""
    model.solver = opt_solver
    summary = model.summary()
    summary.to_string(**kwargs)


@pytest.mark.parametrize(
    "kwargs", [{}, {"names": True}, {"float_format": ".1f"}, {"threshold": 1e-2}]
)
def test_model_summary_to_html(model, opt_solver, kwargs):
    """Test that the summary's method ``to_html`` can be called."""
    model.solver = opt_solver
    summary = model.summary()
    summary.to_html(**kwargs)


def test_model_summary_to_frame_previous_solution(model, opt_solver):
    """Test that the summary correctly uses an existing solution."""
    model.solver = opt_solver
    solution = pfba(model)
    rxn = model.reactions.EX_glc__D_e
    solution.fluxes[rxn.id] = -123
    summary = model.summary(solution=solution)
    assert summary.to_frame().at[rxn.id, "flux"] == 123


def test_model_summary_flux(model, opt_solver):
    """Test that the summary has expected fluxes."""
    model.solver = opt_solver
    summary = model.summary()
    assert summary.uptake_flux.at["EX_o2_e", "flux"] == pytest.approx(21.8, abs=1e-02)
    assert summary.secretion_flux.at["EX_h2o_e", "flux"] == pytest.approx(
        -29.18, abs=1e-02
    )


def test_model_summary_fva(model, opt_solver):
    """Test that the exchange summary is within expected bounds."""
    model.solver = opt_solver
    summary = model.summary(fva=0.95)
    assert summary.to_frame().at["EX_o2_e", "flux"] == pytest.approx(21.8, abs=1e-02)
    assert summary.to_frame().at["EX_o2_e", "minimum"] == pytest.approx(19.9, abs=1e-02)
    assert summary.to_frame().at["EX_o2_e", "maximum"] == pytest.approx(
        23.71, abs=1e-02
    )

    assert summary.to_frame().at["EX_h2o_e", "flux"] == pytest.approx(-29.18, abs=1e-02)
    assert summary.to_frame().at["EX_h2o_e", "minimum"] == pytest.approx(
        -30.72, abs=1e-02
    )
    assert summary.to_frame().at["EX_h2o_e", "maximum"] == pytest.approx(-25, abs=1e-02)


def test_model_summary_flux_in_context(model, opt_solver):
    """Test that the model summary inside and outside of a context are equal."""
    model.solver = opt_solver
    copy = model.copy()
    with model:
        model.reactions.EX_glc__D_e.bounds = (0, 1000)
        model.reactions.EX_ac_e.bounds = (-10, 1000)
        context_summary = model.summary()

    copy.reactions.EX_glc__D_e.bounds = (0, 1000)
    copy.reactions.EX_ac_e.bounds = (-10, 1000)
    outside_summary = copy.summary()

    assert context_summary.to_frame()["flux"].values == pytest.approx(
        outside_summary.to_frame()["flux"].values, abs=model.tolerance
    )



"""Test functionalities of boundary type detection functions."""

import logging

import pytest

from cobra.core import Metabolite, Model, Reaction
from cobra.medium import (
    find_boundary_types,
    find_external_compartment,
    is_boundary_type,
)


def test_find_external_compartment_single(model: Model) -> None:
    """Test external compartment identification."""
    # by name
    assert find_external_compartment(model) == "e"
    # from boundary counts
    for m in model.metabolites:
        if m.compartment == "e":
            m.compartment = "outside"
    for r in model.reactions:
        r._compartments = None
        assert find_external_compartment(model) == "outside"
    # names are always right
    model.exchanges[0].reactants[0].compartment = "extracellular"
    assert find_external_compartment(model) == "extracellular"


def test_find_external_compartment_multi(model: Model) -> None:
    """Test multiple external compartment identification."""
    for r in model.reactions:
        r._compartments = None
    model.exchanges[0].reactants[0].compartment = "extracellular"
    # still works due to different boundary numbers
    assert find_external_compartment(model) == "e"
    model.exchanges[1].reactants[0].compartment = "extra cellular"
    model.remove_reactions(model.exchanges)
    # Now fails because same boundary count
    with pytest.raises(RuntimeError):
        find_external_compartment(model)


@pytest.mark.parametrize("compartment", ["C_e", "e0"])
def test_find_external_popular_reconstructions(
    model: Model, compartment, caplog
) -> None:
    """Test some additional id formats."""
    for ex in model.exchanges:
        ex.reactants[0].compartment = compartment
    with caplog.at_level(logging.WARNING):
        external = find_external_compartment(model)
    assert external == compartment
    assert "complete nonsense" not in caplog.text


def test_no_names_or_boundary_reactions(empty_model: Model) -> None:
    """Test absence of name or boundary reactions."""
    with pytest.raises(RuntimeError):
        find_external_compartment(empty_model)


def test_find_boundary_types_exchange(model: Model) -> None:
    """Test boundary type identification for exchanges."""
    ex = model.exchanges
    assert all(r.id.startswith("EX_") for r in ex)
    ex = find_boundary_types(model, "exchange", "e")
    assert all(r.id.startswith("EX_") for r in ex)


def test_find_boundary_types_demand(model: Model) -> None:
    """Test boundary type identification for demands."""
    dm = Reaction("demand")
    model.add_reactions([dm])
    dm.build_reaction_from_string("atp_c ->")
    dm = model.demands
    assert len(dm) == 1
    assert "demand" in [r.id for r in dm]


def test_find_boundary_types_sink(model: Model) -> None:
    """Test boundary type identification for sinks."""
    sn = Reaction("sink")
    model.add_reactions([sn])
    sn.build_reaction_from_string("atp_c <->")
    sn.bounds = -1000, 1000
    sn = model.sinks
    assert len(sn) == 1
    assert "sink" in [r.id for r in sn]


def test_no_boundary_reactions(empty_model: Model) -> None:
    """Test proper identification of no boundary reactions."""
    assert find_boundary_types(empty_model, "e", None) == []


def test_is_boundary_type(model: Model) -> None:
    """Test correct identification of boundary types for reactions."""
    assert not is_boundary_type(model.reactions.ATPM, "exchange", "e")
    model.reactions.ATPM.annotation["sbo"] = "SBO:0000627"
    assert is_boundary_type(model.reactions.ATPM, "exchange", "bla")
    model.reactions.ATPM.annotation["sbo"] = "SBO:0000632"
    assert not is_boundary_type(model.reactions.ATPM, "exchange", "e")


def test_bad_exchange(model: Model) -> None:
    """Test bad exchange reaction identification."""
    with pytest.raises(ValueError):
        m = Metabolite("baddy", compartment="nonsense")
        model.add_boundary(m, type="exchange")
    m = Metabolite("goody", compartment="e")
    rxn = model.add_boundary(m, type="exchange")
    assert isinstance(rxn, Reaction)



"""Test functions of model.py."""

import os
import warnings
from copy import copy, deepcopy
from math import isnan
from typing import List, Tuple

import numpy as np
import pandas as pd
import pytest
from optlang.symbolics import Expr, Zero

from cobra import Solution
from cobra.core import Group, Metabolite, Model, Reaction
from cobra.exceptions import OptimizationError
from cobra.manipulation.delete import remove_genes
from cobra.util import solver as su
from cobra.util.solver import SolverNotFound, set_objective, solvers


try:
    import pytest_benchmark
except ImportError:
    pytest_benchmark = None

if pytest_benchmark:
    from pytest_benchmark.fixture import BenchmarkFixture


stable_optlang = ["glpk", "cplex", "gurobi"]
optlang_solvers = ["optlang-" + s for s in stable_optlang if s in su.solvers]


def same_ex(ex1: Expr, ex2: Expr) -> bool:
    """Compare two sympy-expressions for mathematical equality.

    Parameters
    ----------
    ex1 : optlang.symbolics.Expr
        The first sympy-expression.
    ex2 : optlang.symbolics.Expr
        The second sympy-expression.

    Returns
    -------
    bool
        Whether the two expression are mathematically equal.

    """
    return ex1.simplify() == ex2.simplify()


def test_add_metabolite(model: Model) -> None:
    """Tests adding a metabolite to a model, including with context.

    Parameters
    ----------
    model: cobra.Model
    """
    new_metabolite = Metabolite("test_met")
    assert new_metabolite not in model.metabolites
    with model:
        model.add_metabolites(new_metabolite)
        assert new_metabolite._model == model
        assert new_metabolite in model.metabolites
        assert new_metabolite.id in model.solver.constraints

    assert new_metabolite._model is None
    assert new_metabolite not in model.metabolites
    assert new_metabolite.id not in model.solver.constraints


def test_remove_metabolite_subtractive(model: Model) -> None:
    """Remove metabolite from model in a subtractive (not destructive) way.

    Checks that the changes to model are reversed when using context.

    Parameters
    ----------
    model: cobra.Model
    """
    test_metabolite = model.metabolites[4]
    test_reactions = test_metabolite.reactions
    with model:
        model.remove_metabolites(test_metabolite, destructive=False)
        assert test_metabolite._model is None
        assert test_metabolite not in model.metabolites
        assert test_metabolite.id not in model.solver.constraints
        for reaction in test_reactions:
            assert reaction in model.reactions

    assert test_metabolite._model is model
    assert test_metabolite in model.metabolites
    assert test_metabolite.id in model.solver.constraints


def test_remove_metabolite_destructive(model: Model) -> None:
    """Remove metabolite from a model in a destructive way.

    Checks that the changes to model are reversed when using context.

    Parameters
    ----------
    model: cobra.Model
    """
    test_metabolite = model.metabolites[4]
    test_reactions = test_metabolite.reactions
    with model:
        model.remove_metabolites(test_metabolite, destructive=True)
        assert test_metabolite._model is None
        assert test_metabolite not in model.metabolites
        assert test_metabolite.id not in model.solver.constraints
        for reaction in test_reactions:
            assert reaction not in model.reactions

    assert test_metabolite._model is model
    assert test_metabolite in model.metabolites
    assert test_metabolite.id in model.solver.constraints
    for reaction in test_reactions:
        assert reaction in model.reactions


def test_compartments(model: Model) -> None:
    """Test setting and modifying model compartments.

    Parameters
    ----------
    model: cobra.Model
    """
    assert set(model.compartments) == {"c", "e"}
    model = Model("test", "test")
    met_c = Metabolite("a_c", compartment="c")
    met_e = Metabolite("a_e", compartment="e")
    rxn = Reaction("foo")
    rxn.add_metabolites({met_e: -1, met_c: 1})
    model.add_reactions([rxn])
    assert model.compartments == {"c": "", "e": ""}
    model.compartments = {"c": "cytosol"}
    assert model.compartments == {"c": "cytosol", "e": ""}


def test_model_remove_reaction(model: Model) -> None:
    """Test remove_reactions() to remove reaction(s).

    Parameters
    ----------
    model: cobra.Model

    """
    old_reaction_count = len(model.reactions)

    with model:
        model.remove_reactions(["PGI"])
        assert len(model.reactions) == old_reaction_count - 1
        with pytest.raises(KeyError):
            model.reactions.get_by_id("PGI")
        model.remove_reactions(model.reactions[:1])
        assert len(model.reactions) == old_reaction_count - 2

    assert len(model.reactions) == old_reaction_count
    assert "PGI" in model.reactions

    tmp_metabolite = Metabolite("testing")
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert tmp_metabolite in model.metabolites
    model.remove_reactions(model.reactions[:1], remove_orphans=True)
    assert tmp_metabolite not in model.metabolites

    with model:
        model.reactions[0].add_metabolites({tmp_metabolite: 1})
        assert tmp_metabolite in model.metabolites
    assert tmp_metabolite not in model.metabolites

    biomass_before = model.slim_optimize()
    with model:
        model.remove_reactions([model.reactions.Biomass_Ecoli_core])
        assert np.isclose(model.slim_optimize(), 0)

    assert np.isclose(model.slim_optimize(), biomass_before)


def test_reaction_remove(model: Model) -> None:
    """Test remove orphans in Reaction().remove_from_model.

    This function test that remove_orphans=True removes related metabolites when
    supposed to and doesn't remove when not supposed to (when this metabolite is in
    reactions not removed).

    Parameters
    ----------
    model: cobra.Model to use
    """
    old_reaction_count = len(model.reactions)
    tmp_metabolite = Metabolite("testing")

    # Delete without removing orphan
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 1

    # Esnsure the stoichiometry is still the same using different objects
    removed_reaction = model.reactions[0]
    original_stoich = {
        i.id: value for i, value in removed_reaction._metabolites.items()
    }
    model.reactions[0].remove_from_model(remove_orphans=False)
    assert len(original_stoich) == len(removed_reaction._metabolites)
    for met in removed_reaction._metabolites:
        assert original_stoich[met.id] == removed_reaction._metabolites[met]
        assert met is not model.metabolites

    # Make sure it's still in the model
    assert tmp_metabolite in model.metabolites
    assert len(tmp_metabolite.reactions) == 0
    assert len(model.reactions) == old_reaction_count - 1

    # Now try with removing orphans
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 1
    model.reactions[0].remove_from_model(remove_orphans=True)
    assert tmp_metabolite not in model.metabolites
    assert len(tmp_metabolite.reactions) == 0
    assert len(model.reactions) == old_reaction_count - 2

    # It shouldn't remove orphans if it's in 2 reactions however
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    model.reactions[1].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 2
    model.reactions[0].remove_from_model(remove_orphans=False)
    assert tmp_metabolite in model.metabolites
    assert len(tmp_metabolite.reactions) == 1
    assert len(model.reactions) == old_reaction_count - 3


def test_reaction_delete(model: Model) -> None:
    """Test reaction removal using the Reaction.delete() function.

    This function calls Reaction().remove_from_model, since it is deprecated.

    Parameters
    ----------
    model: cobra.Model to use
    """
    old_reaction_count = len(model.reactions)
    tmp_metabolite = Metabolite("testing")

    # Delete without removing orphan
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 1
    with pytest.warns(DeprecationWarning):
        model.reactions[0].delete(remove_orphans=False)

    # Make sure it's still in the model
    assert tmp_metabolite in model.metabolites
    assert len(tmp_metabolite.reactions) == 0
    assert len(model.reactions) == old_reaction_count - 1

    # Now try it with removing orphans
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 1
    model.reactions[0].delete(remove_orphans=True)
    assert tmp_metabolite not in model.metabolites
    assert len(tmp_metabolite.reactions) == 0
    assert len(model.reactions) == old_reaction_count - 2

    # It shouldn't remove orphans if it's in 2 reactions however
    model.reactions[0].add_metabolites({tmp_metabolite: 1})
    model.reactions[1].add_metabolites({tmp_metabolite: 1})
    assert len(tmp_metabolite.reactions) == 2
    model.reactions[0].delete(remove_orphans=False)
    assert tmp_metabolite in model.metabolites
    assert len(tmp_metabolite.reactions) == 1
    assert len(model.reactions) == old_reaction_count - 3


def test_remove_gene(model: Model) -> None:
    """Test remove_gene from model.

    Parameters
    ----------
    model: cobra.Model
    """
    target_gene = model.genes[0]
    gene_reactions = list(target_gene.reactions)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        remove_genes(model, [target_gene])
    assert target_gene.model is None

    # Make sure the reaction was removed from the model
    assert target_gene not in model.genes

    # Ensure the old reactions no longer have a record of the gene
    for reaction in gene_reactions:
        assert target_gene not in reaction.genes


def test_group_model_reaction_association(model: Model) -> None:
    """Test associating reactions with group in a model.

    This function will also remove the group from the model and check that
    reactions are no longer associated with the group

    Parameters
    ----------
    model: cobra.Model
    """
    num_members = 5
    reactions_for_group = model.reactions[0:num_members]
    group = Group("arbitrary_group1")
    group.add_members(reactions_for_group)
    group.kind = "collection"
    model.add_groups([group])
    # group should point to and be associated with the model
    assert group._model is model
    assert group in model.groups

    # model.get_associated_groups should find the group for each reaction
    # we added to the group
    for reaction in reactions_for_group:
        assert group in model.get_associated_groups(reaction)

    # remove the group from the model
    model.remove_groups([group])
    assert group not in model.groups
    assert group._model is not model
    for reaction in reactions_for_group:
        assert group not in model.get_associated_groups(reaction)


def test_group_members_add_to_model(model: Model) -> None:
    """Test adding a group with reactions to a model.

    This function will remove some reactions from a model, add them to a group, and
    test that the reactions aren't in the model. Later, the test will add the group
    to the model, and check the reactions were added to the model.

    Parameters
    ----------
    model: cobra.Model
    """
    # remove a few reactions from the model and add them to a new group
    num_members = 5
    reactions_for_group = model.reactions[0:num_members]
    model.remove_reactions(reactions_for_group, remove_orphans=False)
    group = Group("arbitrary_group1")
    group.add_members(reactions_for_group)
    group.kind = "collection"
    # the old reactions should not be in the model
    for reaction in reactions_for_group:
        assert reaction not in model.reactions

    # add the group to the model and check that the reactions were added
    model.add_groups([group])
    assert group in model.groups
    for reaction in reactions_for_group:
        assert reaction in model.reactions


def test_group_loss_of_elements(model: Model) -> None:
    """Test removal from model removes elements from group.

    This function will test that when a metabolite, reaction or gene is removed from
    a model, it no longer is a member of any groups.

    Parameters
    ----------
    model: cobra.Model
    """
    num_members_each = 5
    elements_for_group = model.reactions[0:num_members_each]
    elements_for_group.extend(model.metabolites[0:num_members_each])
    elements_for_group.extend(model.genes[0:num_members_each])
    group = Group("arbitrary_group1")
    group.add_members(elements_for_group)
    group.kind = "collection"
    model.add_groups([group])

    remove_met = model.metabolites[0]
    model.remove_metabolites([remove_met])
    remove_rxn = model.reactions[0]
    model.remove_reactions([remove_rxn])
    remove_gene = model.genes[0]
    remove_genes(model, [remove_gene])
    assert remove_met not in group.members
    assert remove_rxn not in group.members
    assert remove_gene not in group.members


def test_exchange_reactions(model: Model) -> None:
    """Test model.exchanges works as intended.

    Parameters
    ----------
    model: cobra.Model
    """
    assert set(model.exchanges) == {
        rxn for rxn in model.reactions if rxn.id.startswith("EX")
    }


@pytest.mark.parametrize(
    "metabolites, reaction_type, prefix",
    [
        ("exchange", "exchange", "EX_"),
        ("demand", "demand", "DM_"),
        ("sink", "sink", "SK_"),
    ],
    indirect=["metabolites"],
)
def test_add_boundary(
    model: Model, metabolites: List[Metabolite], reaction_type: str, prefix: str
) -> None:
    """Test add_boundary() function for model.

    Parameters
    ----------
    model: cobra.Model
    metabolites: List[Metabolites]
        This list is generated by the pytest.fixture metabolites(), see conftest.py
        in the root test directory.
    reaction_type: {"exchange", "demand", "sink"}
        The allowed types for boundary, see add_boundary() for types.
    prefix: str
    """
    for metabolite in metabolites:
        reaction = model.add_boundary(metabolite, reaction_type)
        assert model.reactions.get_by_id(reaction.id) == reaction
        assert reaction.reactants == [metabolite]
        assert model.constraints[metabolite.id].expression.has(
            model.variables[prefix + metabolite.id]
        )


@pytest.mark.parametrize(
    "metabolites, reaction_type, prefix",
    [
        ("exchange", "exchange", "EX_"),
        ("demand", "demand", "DM_"),
        ("sink", "sink", "SK_"),
    ],
    indirect=["metabolites"],
)
def test_add_boundary_context(
    model: Model, metabolites: List[Metabolite], reaction_type: str, prefix: str
) -> None:
    """Test add_boundary() function for model with context.

    Parameters
    ----------
    model: cobra.Model
    metabolites: List[Metabolites]
        This list is generated by the pytest.fixture metabolites(), see conftest.py
        in the root test directory.
    reaction_type: {"exchange", "demand", "sink"}
        The allowed types for boundary, see add_boundary() for types.
    prefix: str
    """
    with model:
        for metabolite in metabolites:
            reaction = model.add_boundary(metabolite, reaction_type)
            assert model.reactions.get_by_id(reaction.id) == reaction
            assert reaction.reactants == [metabolite]
            assert -model.constraints[metabolite.id].expression.has(
                model.variables[prefix + metabolite.id]
            )
    for metabolite in metabolites:
        assert prefix + metabolite.id not in model.reactions
        assert prefix + metabolite.id not in model.variables.keys()


@pytest.mark.parametrize(
    "metabolites, reaction_type",
    [("exchange", "exchange"), ("demand", "demand"), ("sink", "sink")],
    indirect=["metabolites"],
)
def test_add_existing_boundary(
    model: Model, metabolites: List[Metabolite], reaction_type: str
) -> None:
    """Test add_boundary() function for model with existing boundary/metabolite.

    Parameters
    ----------
    model: cobra.Model
    metabolites: List[Metabolites]
        This list is generated by the pytest.fixture metabolites(), see conftest.py
        in the root test directory.
    reaction_type: {"exchange", "demand", "sink"}
        The allowed types for boundary, see add_boundary() for types.
    """
    for metabolite in metabolites:
        rxn_added = model.add_boundary(metabolite, reaction_type)
        rxn_dup = model.add_boundary(metabolite, reaction_type)
        assert rxn_dup is rxn_added


@pytest.mark.parametrize("solver", optlang_solvers)
def test_copy_benchmark(model: Model, solver: str, benchmark: BenchmarkFixture) -> None:
    """Test copying a model with benchmark.

    Parameters
    ----------
    model: cobra.Model
    solver: str
        It is a string representing which solver to use. Parametized using
        'optlang_solvers' defined above.
    benchmark: BenchmarkFixture

    """

    def _() -> None:
        """Copy a model.

        If the model has no solver, it creates the problem using the solver given as
        parameter to the external function.
        """
        model.solver = solver
        model.copy()

    benchmark(_)


@pytest.mark.parametrize("solver", optlang_solvers)
def test_copy_benchmark_large_model(
    large_model: Model,
    solver: str,
    benchmark: BenchmarkFixture,
) -> None:
    """Test copying a large model with benchmark.

    Parameters
    ----------
    large_model: cobra.Model
    solver: str
        It is a string representing which solver to use. Parametized using
        'optlang_solvers' defined above.
    benchmark: BenchmarkFixture
    """

    def _() -> None:
        """Copy a large model.

        If the model has no solver, it creates the problem using the solver given as
        parameter to the external function.
        """
        large_model.solver = solver
        large_model.copy()

    benchmark(_)


def test_copy(model: Model) -> None:
    """Test copying a model and modifying the copy.

    This function tests that modifying the copy should not modifying the original, by
    deleting reactions in the copy (# of reactions in the original should not change).
    This function also tests that GPRs are copied by content, not by reference, and
    that the model copy does not copy the context.

    Parameters
    ----------
    model: cobra.Model

    """
    # Deleting reactions in copy does not change number of reactions in the original
    model_copy = model.copy()
    old_reaction_count = len(model.reactions)
    assert model_copy.notes is not model.notes
    assert model_copy.annotation is not model.annotation
    assert len(model.reactions) == len(model_copy.reactions)
    assert len(model.metabolites) == len(model_copy.metabolites)
    assert len(model.groups) == len(model_copy.groups)
    assert len(model.genes) == len(model_copy.genes)
    # test if GPRs are copied by content but not by reference
    assert model.reactions[0].gpr == model_copy.reactions[0].gpr
    assert id(model.reactions[0].gpr.body) != id(model_copy.reactions[0].gpr.body)
    model_copy.remove_reactions(model_copy.reactions[0:5])
    assert old_reaction_count == len(model.reactions)
    assert len(model.reactions) != len(model_copy.reactions)
    # Copying a model should not copy its context
    with model:
        model.remove_reactions([model.reactions.ACALD])
        cp_model = model.copy()
        assert len(cp_model._contexts) == 0
    assert "ACALD" not in cp_model.reactions


def test_copy_with_groups(model: Model) -> None:
    """Copy model with groups and check that groups are copied correctly.

    Parameters
    ----------
    model: cobra.Model

    """
    sub = Group("pathway", members=[model.reactions.PFK, model.reactions.FBA])
    model.add_groups([sub])
    model_copy = model.copy()
    assert len(model_copy.groups) == len(model.groups)
    assert len(model_copy.groups.get_by_id("pathway")) == len(
        model.groups.get_by_id("pathway")
    )


def test_deepcopy_benchmark(model: Model, benchmark: BenchmarkFixture) -> None:
    """Benchmark deepcopying a model.

    Parameters
    ----------
    model: cobra.Model
    benchmark: BenchmarkFixture
    """
    benchmark(deepcopy, model)


def test_deepcopy(model: Model) -> None:
    """Test deepcopying works, and maintains reference structures.

    Parameters
    ----------
    model: cobra.Model
    """
    # Reference structures are maintained when deepcopying
    model_copy = deepcopy(model)
    for gene, gene_copy in zip(model.genes, model_copy.genes):
        assert gene.id == gene_copy.id
        reactions = sorted(i.id for i in gene.reactions)
        reactions_copy = sorted(i.id for i in gene_copy.reactions)
        assert reactions == reactions_copy
    for reaction, reaction_copy in zip(model.reactions, model_copy.reactions):
        assert reaction.id == reaction_copy.id
        metabolites = sorted(i.id for i in reaction._metabolites)
        metabolites_copy = sorted(i.id for i in reaction_copy._metabolites)
        assert metabolites == metabolites_copy


def test_add_reaction_orphans(model: Model) -> None:
    """Test orphan behavior when adding reactions.

    Need to verify that no orphan genes or metabolites are contained in reactions
    after adding them to the model.

    Parameters
    ---------
    model: cobra.Model
    """
    model = model.__class__("test")
    model.add_reactions((x.copy() for x in model.reactions))
    genes = []
    metabolites = []
    for x in model.reactions:
        genes.extend(x.genes)
        metabolites.extend(x._metabolites)
    orphan_genes = [x for x in genes if x.model is not model]
    orphan_metabolites = [x for x in metabolites if x.model is not model]
    # Check for dangling genes when running Model.add_reactions
    assert len(orphan_genes) == 0
    # Check for dangling metabolites when running Model.add_reactions
    assert len(orphan_metabolites) == 0


def test_merge_models(model: Model, tiny_toy_model: Model) -> None:
    """Test merging models.

    Parameters
    ----------
    model: cobra.Model
    tiny_toy_model: cobra.Model
    """
    with model, tiny_toy_model:
        # Add some cons/vars to tiny_toy_model for testing merging
        tiny_toy_model.add_reactions([Reaction("EX_glc__D_e")])
        variable = tiny_toy_model.problem.Variable("foo")
        constraint = tiny_toy_model.problem.Constraint(
            variable, ub=0, lb=0, name="constraint"
        )
        tiny_toy_model.add_cons_vars([variable, constraint])

        merged = model.merge(
            tiny_toy_model, inplace=False, objective="sum", prefix_existing="tiny_"
        )
        assert "ex1" in merged.reactions
        assert "ex1" not in model.reactions
        assert merged.reactions.ex1.objective_coefficient == 1
        assert (
            merged.reactions.get_by_id("Biomass_Ecoli_core").objective_coefficient == 1
        )
        assert "tiny_EX_glc__D_e" in merged.reactions
        assert "foo" in merged.variables

        # Test reversible in-place model merging
        with model:
            model.merge(
                tiny_toy_model, inplace=True, objective="left", prefix_existing="tiny_"
            )
            assert "ex1" in model.reactions
            assert "constraint" in model.constraints
            assert "foo" in model.variables
            assert "tiny_EX_glc__D_e" in model.reactions
            assert (
                model.objective.expression.simplify()
                == model.reactions.get_by_id(
                    "Biomass_Ecoli_core"
                ).flux_expression.simplify()
            )
        assert "ex1" not in model.reactions
        assert "constraint" not in model.constraints
        assert "foo" not in model.variables
        assert "tiny_EX_glc__D_e" not in model.reactions


@pytest.mark.parametrize("solver", optlang_solvers)
def test_change_objective_benchmark(
    model: Model, benchmark: BenchmarkFixture, solver: str
) -> None:
    """Benchmark changing objective in model.

    Parameters
    ----------
    model: cobra.Model
    benchmark: BenchmarkFixture
    solver: str
        Solver to use. Parametized using 'optlang_solvers' defined above.
    """
    atpm = model.reactions.get_by_id("ATPM")

    def benchmark_change_objective():
        model.objective = atpm.id
        model.solver = solver

    benchmark(benchmark_change_objective)


def test_get_objective_direction(model: Model) -> None:
    """Test getting objective.

    Parameters
    ----------
    model: cobra.Model
    """
    assert model.objective_direction == "max"
    value = model.slim_optimize()
    assert np.isclose(value, 0.874, 1e-3)


def test_set_objective_direction(model: Model) -> None:
    """Test setting objective.

    Parameters
    ----------
    model: cobra.Model
    """
    with model:
        model.objective_direction = "min"
        assert model.objective_direction == "min"
        value = model.slim_optimize()
        assert value == 0.0
    assert model.objective_direction == "max"


def test_slim_optimize(model: Model) -> None:
    """Test slim_optimize with context.

    Parameters
    ----------
    model: cobra.Model

    """
    with model:
        assert model.slim_optimize() > 0.872
        model.reactions.Biomass_Ecoli_core.lower_bound = 10
        assert isnan(model.slim_optimize())
        with pytest.raises(OptimizationError):
            model.slim_optimize(error_value=None)


@pytest.mark.parametrize("solver", optlang_solvers)
def test_optimize(model: Model, solver: str) -> None:
    """Test optimizing a model.

    Parameters
    ----------
    model: cobra.Model
    solver: str
        Solver to use. Parametized using 'optlang_solvers' defined above.
    """
    model.solver = solver
    with model:
        assert model.optimize().objective_value > 0.872
        model.reactions.Biomass_Ecoli_core.lower_bound = 10
        with pytest.warns(UserWarning):
            model.optimize()
        with pytest.raises(OptimizationError):
            model.optimize(raise_error=True)


def test_change_objective(model: Model) -> None:
    """Test changing objective.

    Parameters
    ----------
    model: cobra.Model
    """
    # Test for correct optimization behavior
    model.optimize()
    assert model.reactions.Biomass_Ecoli_core.x > 0.5
    with model:
        model.objective = model.reactions.EX_etoh_e
        model.optimize()
    assert model.reactions.Biomass_Ecoli_core.x < 0.5
    assert model.reactions.Biomass_Ecoli_core.objective_coefficient == 1
    model.optimize()
    assert model.reactions.Biomass_Ecoli_core.x > 0.5
    # Test changing objective
    biomass = model.reactions.get_by_id("Biomass_Ecoli_core")
    atpm = model.reactions.get_by_id("ATPM")
    model.objective = atpm.id
    assert atpm.objective_coefficient == 1.0
    assert biomass.objective_coefficient == 0.0
    assert su.linear_reaction_coefficients(model) == {atpm: 1.0}
    # Change it back using object itself
    model.objective = biomass
    assert atpm.objective_coefficient == 0.0
    assert biomass.objective_coefficient == 1.0
    # Set both to 1 with a list
    model.objective = [atpm, biomass]
    assert atpm.objective_coefficient == 1.0
    assert biomass.objective_coefficient == 1.0
    # Set both using a dict
    model.objective = {atpm: 0.2, biomass: 0.3}
    assert abs(atpm.objective_coefficient - 0.2) < 10**-9
    assert abs(biomass.objective_coefficient - 0.3) < 10**-9
    # Test setting by index
    model.objective = model.reactions.index(atpm)
    assert su.linear_reaction_coefficients(model) == {atpm: 1.0}
    # Test by setting list of indexes
    model.objective = [model.reactions.index(reaction) for reaction in [atpm, biomass]]
    assert su.linear_reaction_coefficients(model) == {atpm: 1.0, biomass: 1.0}


def test_problem_properties(model: Model) -> None:
    """Test model problem properties.

    Parameters
    ----------
    model: cobra.Model
    """
    new_variable = model.problem.Variable("test_variable")
    new_constraint = model.problem.Constraint(Zero, name="test_constraint", lb=0)
    model.add_cons_vars([new_variable, new_constraint])
    assert "test_variable" in model.variables
    assert "test_constraint" in model.constraints
    model.remove_cons_vars([new_constraint, new_variable])
    assert "test_variable" not in model.variables
    assert "test_constraint" not in model.variables


def test_solution_data_frame(model: Model) -> None:
    """Test that solution is transformed correctly to a Pandas data frame.

    Parameters
    ----------
    model: cobra. Model

    """
    solution = model.optimize().to_frame()
    assert isinstance(solution, pd.DataFrame)
    assert "fluxes" in solution
    assert "reduced_costs" in solution


def test_context_manager(model: Model) -> None:
    """Test that the context manager works.

    Parameters
    ----------
    model: cobra.Model

    """
    bounds0 = model.reactions[0].bounds
    bounds1 = (1, 2)
    bounds2 = (3, 4)

    # Trigger a nested model context, ensuring that bounds are
    # preserved at each level
    with model:
        model.reactions[0].bounds = bounds1
        with model:
            model.reactions[0].bounds = bounds2

            assert model.reactions[0].bounds == bounds2
        assert model.reactions[0].bounds == bounds1
    assert model.reactions[0].bounds == bounds0


def test_objective_coefficient_reflects_changed_objective(model: Model) -> None:
    """Test that changing objectives is reflected in the objectives changing.

    Parameters
    ----------
    model: cobra.Model
    """
    biomass_r = model.reactions.get_by_id("Biomass_Ecoli_core")
    assert biomass_r.objective_coefficient == 1
    model.objective = "PGI"
    assert biomass_r.objective_coefficient == 0
    assert model.reactions.PGI.objective_coefficient == 1


def test_change_objective_through_objective_coefficient(model: Model) -> None:
    """Test that changing the objective coefficients will change the objective.

    Parameters
    ----------
    model: cobra.Model
    """
    biomass_r = model.reactions.get_by_id("Biomass_Ecoli_core")
    pgi = model.reactions.PGI
    pgi.objective_coefficient = 2
    coef_dict = model.objective.expression.as_coefficients_dict()
    # Check that objective has been updated
    assert coef_dict[pgi.forward_variable] == 2.0
    assert coef_dict[pgi.reverse_variable] == -2.0
    # Check that original objective is still in there
    assert coef_dict[biomass_r.forward_variable] == 1.0
    assert coef_dict[biomass_r.reverse_variable] == -1.0


def test_transfer_objective(model: Model) -> None:
    """Test assigning objective from a different mdoel objective.

    Parameters
    ----------
    model: cobra.Model
    """
    new_mod = Model("new model")
    new_mod.add_reactions(model.reactions)
    new_mod.objective = model.objective
    assert {str(x) for x in model.objective.expression.args} == {
        str(x) for x in new_mod.objective.expression.args
    }
    new_mod.slim_optimize()
    assert abs(new_mod.objective.value - 0.874) < 0.001


def test_model_from_other_model(model: Model) -> None:
    """Test creating model from other model.

    Parameters
    ----------
    model: cobra.Model
    """
    model = Model(id_or_model=model)
    for reaction in model.reactions:
        assert reaction == model.reactions.get_by_id(reaction.id)


def test_add_reactions(model: Model) -> None:
    """Test add_reactions() function to add reactions to model.

    Parameters
    ----------
    model: cobra.Model
    """
    r1 = Reaction("r1")
    r1.add_metabolites({Metabolite("A"): -1, Metabolite("B"): 1})
    r1.lower_bound, r1.upper_bound = -999999.0, 999999.0
    r2 = Reaction("r2")
    r2.add_metabolites({Metabolite("A"): -1, Metabolite("C"): 1, Metabolite("D"): 1})
    r2.lower_bound, r2.upper_bound = 0.0, 999999.0
    model.add_reactions([r1, r2])
    r2.objective_coefficient = 3.0
    assert r2.objective_coefficient == 3.0
    assert model.reactions[-2] == r1
    assert model.reactions[-1] == r2
    assert isinstance(model.reactions[-2].reverse_variable, model.problem.Variable)
    coefficients_dict = model.objective.expression.as_coefficients_dict()
    biomass_r = model.reactions.get_by_id("Biomass_Ecoli_core")
    assert coefficients_dict[biomass_r.forward_variable] == 1.0
    assert coefficients_dict[biomass_r.reverse_variable] == -1.0
    assert coefficients_dict[model.reactions.r2.forward_variable] == 3.0
    assert coefficients_dict[model.reactions.r2.reverse_variable] == -3.0


def test_add_reactions_single_existing(model: Model) -> None:
    """Test adding a reaction already present to a model.

    Parameters
    ----------
    model: cobra.Model
    """
    rxn = model.reactions[0]
    r1 = Reaction(rxn.id)
    r1.add_metabolites({Metabolite("A"): -1, Metabolite("B"): 1})
    r1.lower_bound, r1.upper_bound = -999999.0, 999999.0
    model.add_reactions([r1])
    assert rxn in model.reactions
    assert r1 is not model.reactions.get_by_id(rxn.id)


def test_add_reactions_duplicate(model: Model) -> None:
    """Test adding duplicate reactions to a model.

    Parameters
    ----------
    model: cobra.Model
    """
    rxn = model.reactions[0]
    r1 = Reaction("r1")
    r1.add_metabolites({Metabolite("A"): -1, Metabolite("B"): 1})
    r1.lower_bound, r1.upper_bound = -999999.0, 999999.0
    r2 = Reaction(rxn.id)
    r2.add_metabolites({Metabolite("A"): -1, Metabolite("C"): 1, Metabolite("D"): 1})
    model.add_reactions([r1, r2])
    assert r1 in model.reactions
    assert rxn in model.reactions
    assert r2 is not model.reactions.get_by_id(rxn.id)


def test_all_objects_point_to_all_other_correct_objects(model: Model) -> None:
    """Test that objects point to needed other objects.

    This will test that the reaction.genes, reaction.metabolites point to the correct
    genes and metabolites in the model.

    Parameters
    ----------
    model: cobra.Model
    """
    for reaction in model.reactions:
        assert reaction.model == model
        for gene in reaction.genes:
            assert gene == model.genes.get_by_id(gene.id)
            assert gene.model == model
            for reaction2 in gene.reactions:
                assert reaction2.model == model
                assert reaction2 == model.reactions.get_by_id(reaction2.id)

        for metabolite in reaction.metabolites:
            assert metabolite.model == model
            assert metabolite == model.metabolites.get_by_id(metabolite.id)
            for reaction2 in metabolite.reactions:
                assert reaction2.model == model
                assert reaction2 == model.reactions.get_by_id(reaction2.id)


def test_objects_point_to_correct_other_after_copy(model: Model) -> None:
    """Test that objects point to correct other objects after copying a model.

    Parameters
    ----------
    model: cobra.Model
    """
    for reaction in model.reactions:
        assert reaction.model == model
        for gene in reaction.genes:
            assert gene == model.genes.get_by_id(gene.id)
            assert gene.model == model
            for reaction2 in gene.reactions:
                assert reaction2.model == model
                assert reaction2 == model.reactions.get_by_id(reaction2.id)

        for metabolite in reaction.metabolites:
            assert metabolite.model == model
            assert metabolite == model.metabolites.get_by_id(metabolite.id)
            for reaction2 in metabolite.reactions:
                assert reaction2.model == model
                assert reaction2 == model.reactions.get_by_id(reaction2.id)


def test_remove_reactions(model: Model) -> None:
    """Test remove_reactions() from Model.

    Parameters
    ----------
    model: cobra.Model
    """
    reactions_to_remove = model.reactions[10:30]
    assert all([reaction.model is model for reaction in reactions_to_remove])
    assert all(
        [
            model.reactions.get_by_id(reaction.id) == reaction
            for reaction in reactions_to_remove
        ]
    )

    model.remove_reactions(reactions_to_remove)
    assert all([reaction.model is None for reaction in reactions_to_remove])
    for reaction in reactions_to_remove:
        assert reaction.id not in list(model.variables.keys())

    model.add_reactions(reactions_to_remove)
    for reaction in reactions_to_remove:
        assert reaction in model.reactions


def test_objective(model: Model) -> None:
    """Test that objective contains the correct coefficients.

    Parameters
    ----------
    model: cobra.Model
    """
    obj = model.objective
    assert obj.get_linear_coefficients(obj.variables) == {
        model.variables["Biomass_Ecoli_core_reverse_2cdba"]: -1,
        model.variables["Biomass_Ecoli_core"]: 1,
    }
    assert obj.direction == "max"


def test_change_objective_with_context(model: Model) -> None:
    """Test changing objective is reversed with context.

    Parameters
    ----------
    model: cobra.Model
    """
    expression = 1.0 * model.variables["ENO"] + 1.0 * model.variables["PFK"]
    model.objective = model.problem.Objective(expression)
    assert same_ex(model.objective.expression, expression)
    model.objective = "ENO"
    eno_obj = model.problem.Objective(
        model.reactions.ENO.flux_expression, direction="max"
    )
    pfk_obj = model.problem.Objective(
        model.reactions.PFK.flux_expression, direction="max"
    )
    assert same_ex(model.objective.expression, eno_obj.expression)

    with model:
        model.objective = "PFK"
        assert same_ex(model.objective.expression, pfk_obj.expression)
    assert same_ex(model.objective.expression, eno_obj.expression)
    expression = model.objective.expression
    atpm = model.reactions.get_by_id("ATPM")
    biomass = model.reactions.get_by_id("Biomass_Ecoli_core")
    with model:
        model.objective = atpm
    assert same_ex(model.objective.expression, expression)
    with model:
        atpm.objective_coefficient = 1
        biomass.objective_coefficient = 2
    assert same_ex(model.objective.expression, expression)

    with model:
        set_objective(model, model.problem.Objective(atpm.flux_expression))
        assert same_ex(model.objective.expression, atpm.flux_expression)
    assert same_ex(model.objective.expression, expression)

    expression = model.objective.expression
    with model:
        with model:  # Test to make sure nested contexts are OK
            set_objective(model, atpm.flux_expression, additive=True)
            assert same_ex(
                model.objective.expression, expression + atpm.flux_expression
            )
    assert same_ex(model.objective.expression, expression)


def test_set_reaction_objective(model: Model) -> None:
    """Test setting reaction objective.

    Parameters
    ----------
    model: cobra.Model
    """
    model.objective = model.reactions.ACALD
    assert same_ex(
        model.objective.expression,
        1.0 * model.reactions.ACALD.forward_variable
        - 1.0 * model.reactions.ACALD.reverse_variable,
    )


def test_set_reaction_objective_str(model: Model) -> None:
    """Test setting reaction objective using string.

    Parameters
    ----------
    model: cobra.Model
    """
    model.objective = model.reactions.ACALD.id
    assert same_ex(
        model.objective.expression,
        1.0 * model.reactions.ACALD.forward_variable
        - 1.0 * model.reactions.ACALD.reverse_variable,
    )


def test_invalid_objective_raises(model: Model) -> None:
    """Test that an invalid objective will raise appropriate errors.

    Parameters
    ----------
    model: cobra.Model
    """
    with pytest.raises(ValueError):
        model.objective = "This is not a valid objective!"
    with pytest.raises(TypeError):
        model.objective = 3.0


@pytest.mark.skipif("cplex" not in solvers, reason="need cplex")
def test_solver_change(model: Model) -> None:
    """Test changing the solver to cplex.

    This test will be skipped if cplex is not installed and available to python.

    Parameters
    ----------
    model: cobra.Model
    """
    model.solver = "glpk"
    solver_id = id(model.solver)
    problem_id = id(model.solver.problem)
    solution = model.optimize().fluxes
    model.solver = "cplex"
    assert id(model.solver) != solver_id
    assert id(model.solver.problem) != problem_id
    new_solution = model.optimize().fluxes
    assert np.allclose(solution, new_solution, rtol=0, atol=1e-06)


def test_no_change_for_same_solver(model: Model) -> None:
    """Test no change in variables if changing to the same solver as before.

    Parameters
    ----------
    model: cobra.Model
    """
    model.solver = "glpk"
    solver_id = id(model.solver)
    problem_id = id(model.solver.problem)
    model.solver = "glpk"
    assert id(model.solver) == solver_id
    assert id(model.solver.problem) == problem_id


def test_invalid_solver_change_raises(model: Model) -> None:
    """Test changing to an invalid solver will raise SovlerNotFound.

    Parameters
    ----------
    model: cobra.Model
    """
    with pytest.raises(SolverNotFound):
        model.solver = [1, 2, 3]
    with pytest.raises(SolverNotFound):
        model.solver = "ThisIsDefinitelyNotAvalidSolver"
    with pytest.raises(SolverNotFound):
        model.solver = os


@pytest.mark.skipif("cplex" not in solvers, reason="no cplex")
def test_change_solver_to_cplex_and_check_copy_works(model: Model) -> None:
    """Test changing solver and copying model work.

    This test will be skipped if cplex is not installed and available for python.

    Parameters
    ----------
    model: cobra.Model
    """
    assert (model.slim_optimize() - 0.8739215069684306) == pytest.approx(0.0)
    model_copy = model.copy()
    assert (model_copy.slim_optimize() - 0.8739215069684306) == pytest.approx(0.0)
    # Second, change existing glpk based model to cplex
    model.solver = "cplex"
    assert (model.slim_optimize() - 0.8739215069684306) == pytest.approx(0.0)
    model_copy = copy(model)
    assert (model_copy.slim_optimize() - 0.8739215069684306) == pytest.approx(0.0)


def test_copy_preserves_existing_solution(solved_model: Tuple[Solution, Model]) -> None:
    """Test copy keeps the existing solution.

    Primal values are the same when copying a solved model.

    Parameters
    ----------
    solved_model: Tuple
        A Tuple that contains a Solution and a Model
    """
    solution, model = solved_model
    model_cp = copy(model)
    primals_original = [variable.primal for variable in model.variables]
    primals_copy = [variable.primal for variable in model_cp.variables]
    abs_diff = abs(np.array(primals_copy) - np.array(primals_original))
    assert not any(abs_diff > 1e-6)


def test_repr_html_(model: Model) -> None:
    """Test HTML representation of model.

    Parameters
    ----------
    model: cobra.Model
    """
    assert "<table>" in model._repr_html_()



"""Test functionalities of I/O in MATLAB (.mat) format."""

from pathlib import Path
from pickle import load
from typing import TYPE_CHECKING, Callable

import pytest

from cobra.io import load_matlab_model, read_sbml_model, save_matlab_model


try:
    import scipy
except ImportError:
    scipy = None


if TYPE_CHECKING:
    from cobra import Model


@pytest.fixture(scope="function")
def raven_model(data_directory: Path) -> "Model":
    """Fixture for RAVEN model."""
    with open(data_directory / "raven.pickle", "rb") as infile:
        return load(infile)


@pytest.mark.skipif(scipy is None, reason="scipy unavailable")
@pytest.mark.parametrize(
    "fix_ref_model, fix_directory, filename",
    [
        ("mini_model", "cobra_data_directory", "mini.mat"),
        ("raven_model", "data_directory", "raven.mat"),
    ],
)
def test_load_matlab_model(
    compare_models: Callable,
    fix_ref_model: str,
    fix_directory: str,
    filename: str,
    request: pytest.FixtureRequest,
) -> None:
    """Test the reading of MAT model.

    Will check Path and str for each model.

    Parameters
    ----------
    compare_models : Callable
        A callable function to compare models.
    fix_ref_model: str
        Name of reference model fixture which will be requested.
    fix_directory: str
        Name of directory fixture which will be requested.
    filename: str
        Filename to use as a parameter
    request: FixtureRequest
        Will be used to request fixtures.

    """
    current_model = load_matlab_model(
        (request.getfixturevalue(fix_directory) / filename).resolve()
    )
    assert compare_models(request.getfixturevalue(fix_ref_model), current_model) is None
    current_model = load_matlab_model(
        str((request.getfixturevalue(fix_directory) / filename).resolve())
    )
    assert compare_models(request.getfixturevalue(fix_ref_model), current_model) is None
    with request.getfixturevalue(fix_directory).joinpath(filename).open("rb") as file_h:
        current_model = load_matlab_model(file_h)
        assert (
            compare_models(request.getfixturevalue(fix_ref_model), current_model)
            is None
        )


# @pytest.mark.xfail(reason="localPath not supported yet")
@pytest.mark.skipif(scipy is None, reason="scipy unavailable")
@pytest.mark.parametrize(
    "fix_ref_model, filename",
    [("mini_model", "mini.mat"), ("raven_model", "raven.mat")],
)
def test_save_matlab_model(tmp_path: Path, fix_ref_model, filename, request) -> None:
    """Test the writing of MAT model.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the temporary test assets store.
    fix_ref_model: str
        Name of reference model fixture which will be requested.
    filename: str
        Filename to use as a parameter
    request: FixtureRequest
        Will be used to request fixtures.
    """
    output_file = tmp_path / filename
    save_matlab_model(
        request.getfixturevalue(fix_ref_model), str(output_file.resolve())
    )
    assert output_file.exists()
    output_file.unlink()
    save_matlab_model(request.getfixturevalue(fix_ref_model), output_file.resolve())
    assert output_file.exists()
    output_file.unlink()
    with output_file.open("wb") as file_h:
        save_matlab_model(request.getfixturevalue(fix_ref_model), file_h)
    assert output_file.exists()


@pytest.mark.skipif(scipy is None, reason="scipy unavailable")
def test_large_bounds(tmp_path: Path, model: "Model") -> None:
    """Verify that mat bounds don't get broken by the config defaults.

    Parameters
    ----------
    tmp_path : pathlib.Path
        The path to the temporary test assets store.
    model : cobra.Model
        The "textbook" model.

    """
    model.reactions[0].bounds = -1e6, 1e6
    filepath = tmp_path / "model.mat"
    save_matlab_model(model, filepath.resolve())
    read = load_matlab_model(filepath.resolve())
    assert read.reactions[0].bounds == (-1e6, 1e6)


@pytest.mark.skipif(scipy is None, reason="scipy unavailable")
@pytest.mark.parametrize(
    "fix_directory, filename",
    [("cobra_data_directory", "mini.mat"), ("data_directory", "raven.mat")],
)
def test_read_rewrite_matlab_model(
    compare_models: Callable,
    tmp_path: Path,
    fix_directory: str,
    filename: str,
    request: pytest.FixtureRequest,
) -> None:
    """Verify that rewritten matlab model is identical to original.

    Parameters
    ----------
    compare_models : Callable
        A callable to compare models.
    tmp_path : pathlib.Path
        The path to the temporary test assets store.
    fix_directory: str
        Name of directory fixture which will be requested.
    filename: str
        Filename to use as a parameter
    request: FixtureRequest
        Will be used to request fixtures.

    """
    current_model = load_matlab_model(request.getfixturevalue(fix_directory) / filename)
    output_file = tmp_path.joinpath(filename)
    save_matlab_model(current_model, output_file)
    mat_model_reload = load_matlab_model(output_file)
    assert compare_models(current_model, mat_model_reload) is None


@pytest.mark.skipif(scipy is None, reason="scipy unavailable")
@pytest.mark.parametrize(
    "dirname, xml_file",
    [
        ("data_directory", "e_coli_core.xml"),
        ("cobra_data_directory", "salmonella.xml.gz"),
        ("cobra_data_directory", "mini_cobra.xml"),
        ("data_directory", "mini_fbc2.xml"),
    ],
)
# When using a better comparison function, can run test on
# "annotation.xml", "example_notes.xml", "fbc_ex1.xml", "fbc_ex2.xml", "validation.xml"
# "valid_annotation_output.xml" has reaction annotations in a metabolite, so they would
# be thrown out by matlab
def test_compare_xml_to_written_matlab_model(
    compare_models: Callable,
    tmp_path: Path,
    dirname: str,
    xml_file: str,
    request: pytest.FixtureRequest,
) -> None:
    """Verify that xml rewritten as mat file is written and read correctly.

    Parameters
    ----------
    compare_models : Callable
        A callable to compare models.
    tmp_path : pathlib.Path
        The path to the temporary test assets store.
    xml_file : str
        The name of the XML file to compare against.

    """
    xml_model = read_sbml_model(request.getfixturevalue(dirname) / xml_file)
    mat_output_file = tmp_path / xml_file.replace(".xml", ".mat")
    save_matlab_model(xml_model, mat_output_file)
    mat_model = load_matlab_model(mat_output_file)
    assert compare_models(xml_model, mat_model) is None


@pytest.mark.skipif(scipy is None, reason="scipy unavailable")
def test_fail_on_problematic_compartments(data_directory: Path) -> None:
    """Test that mat import will fail if there are problems in compartments.

    Parameters
    ----------
    data_directory : pathlib.Path
        The path to the test data directory.

    """
    with pytest.raises(IOError):
        # AntCore does not have defined compartments
        load_matlab_model(data_directory / "AntCore.mat")
    with pytest.raises(IOError):
        # Ec_iAF1260_flux1 has underscore in compartment names which is not allowed
        load_matlab_model(data_directory / "Ec_iAF1260_flux1.mat")


@pytest.mark.skipif(scipy is None, reason="scipy unavailable")
def test_mat_model_with_long_compartment_ids(
    compare_models: Callable, data_directory: Path, tmp_path: Path
) -> None:
    """Test that long compartment IDs like "luSI" are correctly loaded.

    Parameters
    ----------
    compare_models : Callable
        A callable to compare models.
    data_directory : pathlib.Path
        The path to the test data directory.
    tmp_path : pathlib.Path
        The path to the temporary test assets store.

    """
    model_compartments = load_matlab_model(data_directory / "compartments.mat")
    assert model_compartments.compartments == {
        "csf": "csf",
        "bcK": "bcK",
        "a": "a",
        "luSI": "luSI",
        "luLI": "luLI",
        "luP": "luP",
        "aL": "aL",
        "fe": "fe",
    }
    assert len(model_compartments.metabolites) == 8
    assert len(model_compartments.reactions) == 15
    for met in model_compartments.metabolites:
        assert met.annotation == {
            "bigg.metabolite": ["glc__D"],
            "cas": ["50-99-7"],
            "kegg.compound": ["C00031"],
            "pubchem.substance": ["3333"],
        }
    output_file = tmp_path / "compartments.mat"
    save_matlab_model(model_compartments, output_file.resolve())
    model_compartments_reloaded = load_matlab_model(output_file.resolve())
    assert compare_models(model_compartments, model_compartments_reloaded) is None


@pytest.mark.skipif(scipy is None, reason="scipy unavailable")
def test_mat_model_with_no_genes(
    compare_models: Callable, data_directory: Path, tmp_path: Path
) -> None:
    """Test that a model with no genes is loaded and reloaded correctly.

    Parameters
    ----------
    compare_models : Callable
        A callable to compare models.
    data_directory : pathlib.Path
        The path to the test data directory.
    tmp_path : pathlib.Path
        The path to the temporary test assets store.

    """
    model_no_genes = load_matlab_model(
        data_directory / "cardiac_mit_glcuptake_atpmax.mat"
    )
    assert not len(model_no_genes.genes)
    output_file = tmp_path / "cardiac_mit_glcuptake_atpmax.mat"
    save_matlab_model(model_no_genes, output_file.resolve())
    model_no_genes_reloaded = load_matlab_model(output_file.resolve())
    assert compare_models(model_no_genes, model_no_genes_reloaded) is None


@pytest.mark.skipif(scipy is None, reason="scipy unavailable")
def test_mat_model_wrong_caps(
    compare_models: Callable, data_directory: Path, cobra_data_directory: Path
) -> None:
    """Check that wrong capitalization in matlab field names is processed correctly.

    See https://gist.github.com/akaviaLab/3dcb0eed6563a9d3d1e07198337300ac to create it
    again when needed.

    Parameters
    ----------
    compare_models : Callable
        A callable to compare models.
    data_directory : pathlib.Path
        The path to the test data directory.

    """
    mat_model = load_matlab_model((cobra_data_directory / "mini.mat").resolve())
    mat_wrong_caps_model = load_matlab_model(
        (data_directory / "mini_wrong_key_caps.mat").resolve()
    )
    assert compare_models(mat_model, mat_wrong_caps_model) is None
    assert mat_wrong_caps_model.reactions.get_by_id("LDH_D").annotation == {
        "rhea": ["16369", "16370", "16371", "16372"],
        "metanetx.reaction": ["MNXR101037"],
        "kegg.reaction": ["R00704"],
        "bigg.reaction": ["LDH_D"],
        "ec-code": ["1.1.1.28"],
        "biocyc": ["META:DLACTDEHYDROGNAD-RXN"],
        "sbo": ["SBO:0000375"],
    }
    for rxn in mat_model.reactions.list_attr("id"):
        assert (
            mat_wrong_caps_model.reactions.get_by_id(rxn).annotation
            == mat_model.reactions.get_by_id(rxn).annotation
        )
    assert mat_wrong_caps_model.metabolites.get_by_id("pyr_c").annotation == {
        "seed.compound": ["cpd00020"],
        "unipathway.compound": ["UPC00022"],
        "lipidmaps": ["LMFA01060077"],
        "reactome": ["REACT_113557", "REACT_389680", "REACT_29398"],
        "biocyc": ["PYRUVATE"],
        "chebi": [
            "CHEBI:15361",
            "CHEBI:14987",
            "CHEBI:8685",
            "CHEBI:32816",
            "CHEBI:45253",
            "CHEBI:26466",
            "CHEBI:26462",
        ],
        "pubchem.substance": ["3324"],
        "bigg.metabolite": ["pyr"],
        "cas": ["127-17-3"],
        "hmdb": ["HMDB00243"],
        "kegg.compound": ["C00022"],
    }
    for met in mat_model.metabolites.list_attr("id"):
        assert (
            mat_wrong_caps_model.metabolites.get_by_id(met).annotation
            == mat_model.metabolites.get_by_id(met).annotation
        )



"""Test functionalities of I/O in JSON format."""

import json
from importlib.resources import files
from pathlib import Path
from typing import Any, Callable, Dict, Union

import pytest

from cobra import Model
from cobra import io as cio


@pytest.fixture(scope="module")
def json_schema_v1() -> Dict[str, Union[str, bool, Any]]:
    """Fixture for cobra JSON-schema."""
    with files(cio).joinpath("schema_v1.json").open("r") as handle:
        schema_v1 = json.load(handle)
    return schema_v1


def test_validate_json(
    cobra_data_directory: Path, json_schema_v1: Dict[str, Union[str, bool, Any]]
) -> None:
    """Validate file according to JSON-schema."""
    jsonschema = pytest.importorskip("jsonschema")
    with open(
        cobra_data_directory.joinpath("mini.json"), "r", encoding="utf-8"
    ) as infile:
        loaded = json.load(infile)
    assert jsonschema.validate(loaded, json_schema_v1) is None


def test_load_json_model(
    compare_models: Callable, cobra_data_directory: Path, mini_model: Model
) -> None:
    """Test the reading of JSON model."""
    json_model = cio.load_json_model(cobra_data_directory / "mini.json")
    assert compare_models(mini_model, json_model) is None
    json_model = cio.load_json_model(str(cobra_data_directory / "mini.json"))
    assert compare_models(mini_model, json_model) is None
    with open(cobra_data_directory / "mini.json", "r") as json_handle:
        json_model = cio.load_json_model(json_handle)
        assert compare_models(mini_model, json_model) is None


def test_save_json_model(
    tmp_path: Path,
    mini_model: Model,
    json_schema_v1: Dict[str, Union[str, bool, Any]],
) -> None:
    """Test the writing of JSON model."""
    jsonschema = pytest.importorskip("jsonschema")
    output_file = tmp_path.joinpath("mini.json")
    cio.save_json_model(mini_model, output_file, pretty=True)
    # validate against JSONSchema
    with open(output_file, "r") as infile:
        loaded = json.load(infile)
    assert jsonschema.validate(loaded, json_schema_v1) is None
    output_file.unlink()
    cio.save_json_model(mini_model, str(output_file), pretty=True)
    # validate against JSONSchema
    with open(output_file, "r") as infile:
        loaded = json.load(infile)
    assert jsonschema.validate(loaded, json_schema_v1) is None
    output_file.unlink()
    with output_file.open("w+") as json_outfile:
        cio.save_json_model(mini_model, json_outfile, pretty=True)
        # validate against JSONSchema
        json_outfile.seek(0, 0)
        loaded = json.load(json_outfile)
        assert jsonschema.validate(loaded, json_schema_v1) is None


def test_reaction_bounds_json(data_directory: Path, tmp_path: Path) -> None:
    """Test reading and writing of model with inf bounds in JSON."""
    # Path to XML file with INF bounds
    path_to_xml_inf_file = data_directory / "fbc_ex1.xml"
    model_xml_inf = cio.read_sbml_model(path_to_xml_inf_file)
    path_to_output = tmp_path.joinpath("fbc_ex1_json.json")
    # Saving model with inf bounds in json form without error
    cio.save_json_model(model_xml_inf, path_to_output)
    # Path to JSON file with INF bounds
    path_to_JSON_inf_file = data_directory.joinpath("JSON_with_inf_bounds.json")
    model_json_inf = cio.load_json_model(path_to_JSON_inf_file)
    assert model_json_inf.reactions[0].upper_bound == float("inf")



"""Test functionalities of flux sampling methods."""

import os

import numpy as np
import pytest

from cobra.core import Metabolite, Model, Reaction
from cobra.flux_analysis.parsimonious import pfba
from cobra.sampling import ACHRSampler, OptGPSampler, sample


def test_single_achr(model: Model) -> None:
    """Test ACHR sampling (one sample)."""
    s = sample(model, 10, method="achr")
    assert s.shape == (10, len(model.reactions))


def test_single_optgp(model: Model) -> None:
    """Test OptGP sampling (one sample)."""
    s = sample(model, 10, processes=1)
    assert s.shape == (10, len(model.reactions))


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_multi_optgp(model: Model) -> None:  # pragma: no cover
    """Test OptGP sampling (multi sample)."""
    s = sample(model, 10, processes=2)
    assert s.shape == (10, len(model.reactions))


def test_wrong_method(model: Model) -> None:
    """Test method intake sanity."""
    with pytest.raises(ValueError):
        sample(model, 1, method="schwupdiwupp")


def test_fixed_seed(model: Model) -> None:
    """Test result of fixed seed for sampling."""
    s1 = sample(model, 1, seed=42)
    s2 = sample(model, 1, seed=42)
    assert np.isclose(s1.TPI[0], s2.TPI[0])


def test_equality_constraint(model: Model) -> None:
    """Test equality constraint."""
    model.reactions.ACALD.bounds = (-1.5, -1.5)

    s = sample(model, 10)
    assert np.allclose(s.ACALD, -1.5, atol=1e-6, rtol=0)

    s = sample(model, 10, method="achr")
    assert np.allclose(s.ACALD, -1.5, atol=1e-6, rtol=0)


def test_inequality_constraint(model: Model) -> None:
    """Test inequality constraint."""
    co = model.problem.Constraint(model.reactions.ACALD.flux_expression, lb=-0.5)
    model.add_cons_vars(co)

    s = sample(model, 10)
    assert all(s.ACALD > -0.5 - 1e-6)

    s = sample(model, 10, method="achr")
    assert all(s.ACALD > -0.5 - 1e-6)


def test_inhomogeneous_sanity(model: Model) -> None:
    """Test standard deviation between inhomogeneous and homogeneous sampling."""
    model.reactions.ACALD.bounds = (-1.5, -1.5)
    s_inhom = sample(model, 64)

    model.reactions.ACALD.bounds = (-1.5 - 1e-3, -1.5 + 1e-3)
    s_hom = sample(model, 64)

    relative_diff = (s_inhom.std() + 1e-12) / (s_hom.std() + 1e-12)
    assert 0.5 < relative_diff.abs().mean() < 2

    model.reactions.ACALD.bounds = (-1.5, -1.5)
    s_inhom = sample(model, 64, method="achr")

    model.reactions.ACALD.bounds = (-1.5 - 1e-3, -1.5 + 1e-3)
    s_hom = sample(model, 64, method="achr")

    relative_diff = (s_inhom.std() + 1e-12) / (s_hom.std() + 1e-12)
    assert 0.5 < relative_diff.abs().mean() < 2


def test_complicated_model() -> None:
    """Test a complicated model.

    Difficult model since the online mean calculation is numerically
    unstable, so many samples weakly violate the equality constraints.

    """
    model = Model("flux_split")

    reaction1 = Reaction("V1")
    reaction2 = Reaction("V2")
    reaction3 = Reaction("V3")
    reaction1.bounds = (0, 6)
    reaction2.bounds = (0, 8)
    reaction3.bounds = (0, 10)

    A = Metabolite("A")

    reaction1.add_metabolites({A: -1})
    reaction2.add_metabolites({A: -1})
    reaction3.add_metabolites({A: 1})

    model.add_reactions([reaction1, reaction2, reaction3])

    optgp = OptGPSampler(model, 1, seed=42)
    achr = ACHRSampler(model, seed=42)

    optgp_samples = optgp.sample(100)
    achr_samples = achr.sample(100)

    assert any(optgp_samples.corr().abs() < 1.0)
    assert any(achr_samples.corr().abs() < 1.0)

    # > 95% are valid
    assert sum(optgp.validate(optgp_samples) == "v") > 95
    assert sum(achr.validate(achr_samples) == "v") > 95


def test_single_point_space(model: Model) -> None:
    """Test the reduction of the sampling space to one point."""
    pfba_sol = pfba(model)
    pfba_const = model.problem.Constraint(
        sum(model.variables), ub=pfba_sol.objective_value
    )
    model.add_cons_vars(pfba_const)
    model.reactions.Biomass_Ecoli_core.lower_bound = pfba_sol.fluxes.Biomass_Ecoli_core

    with pytest.raises(ValueError):
        sample(model, 1)



"""Test functionalities of minimal medium creation and analysis."""

from typing import Callable

import pandas as pd
import pytest

from cobra.core import Model, Reaction
from cobra.medium import minimal_medium


def test_minimal_medium_linear(model: Model) -> None:
    """Test linear minimal medium."""
    med = minimal_medium(model, 0.8)
    assert len(med) <= 4
    assert all(med > 1e-6)


def test_minimal_medium_linear_benchmark(model: Model, benchmark: Callable) -> None:
    """Benchmark linear minimal medium."""
    benchmark(minimal_medium, model, 0.8)


def test_minimal_medium_mip(model: Model) -> None:
    """Test mixed-integer minimal medium."""
    med = minimal_medium(model, 0.8, minimize_components=True)
    assert len(med) <= 4
    assert all(med > 1e-6)
    # Anaerobic growth
    med = minimal_medium(model, 0.1, minimize_components=True)
    assert len(med) <= 3
    assert all(med > 1e-6)


def test_minimal_medium_mip_benchmark(model: Model, benchmark: Callable) -> None:
    """Benchmark mixed-integer minimal medium."""
    benchmark(minimal_medium, model, 0.8, True)


def test_minimal_medium_alternative_mip(model: Model) -> None:
    """Test alternative mixed-integer minimal medium."""
    med = minimal_medium(model, 0.8, minimize_components=5, open_exchanges=True)
    assert isinstance(med, pd.DataFrame)
    assert med.shape[0] >= 5
    assert med.shape[1] == 5
    assert all((med > 0).sum() == 3)
    assert all(med.sum(axis=1) > 1e-6)


def test_minimal_medium_exports(model: Model) -> None:
    """Test exports of a minimal medium."""
    med = minimal_medium(model, 0.8, exports=True, minimize_components=True)
    assert len(med) > 4
    assert any(med < -1e-6)


def test_minimal_medium_open_exchanges(model: Model) -> None:
    """Test open exchanges of a minimal medium."""
    model.reactions.EX_glc__D_e.bounds = 0, 0
    med = minimal_medium(model, 0.8)
    assert med is None
    med = minimal_medium(model, 0.8, minimize_components=True)
    assert med is None

    med = minimal_medium(model, 0.8, open_exchanges=True)
    assert len(med) >= 3
    med = minimal_medium(model, 0.8, open_exchanges=100)
    assert len(med) >= 3


def test_model_medium(model: Model) -> None:
    """Test proper functioning of model medium manipulations."""
    # Add a dummy 'malformed' import reaction
    bad_import = Reaction("bad_import")
    bad_import.add_metabolites({model.metabolites.pyr_c: 1})
    bad_import.bounds = (0, 42)
    model.add_reactions([bad_import])

    # Test basic setting and getting methods
    medium = model.medium
    model.medium = medium
    assert model.medium == medium

    # Test context management
    with model:
        # Ensure the bounds are correct beforehand
        assert model.reactions.EX_glc__D_e.lower_bound == -10
        assert model.reactions.bad_import.upper_bound == 42
        assert model.reactions.EX_co2_e.lower_bound == -1000

        # Make changes to the media
        new_medium = model.medium
        new_medium["EX_glc__D_e"] = 20
        new_medium["bad_import"] = 24
        del new_medium["EX_co2_e"]

        # Change the medium, make sure changes work
        model.medium = new_medium
        assert model.reactions.EX_glc__D_e.lower_bound == -20
        assert model.reactions.bad_import.upper_bound == 24
        assert model.reactions.EX_co2_e.lower_bound == 0

    # Make sure changes revert after the contex
    assert model.reactions.EX_glc__D_e.lower_bound == -10
    assert model.reactions.bad_import.upper_bound == 42
    assert model.reactions.EX_co2_e.lower_bound == -1000

    new_medium["bogus_rxn"] = 0
    with pytest.raises(KeyError):
        model.medium = new_medium


def test_medium_does_not_affect_reactant_exports(model: Model) -> None:
    """Test that the medium setter does not overwrite exports defined as reactants."""
    med = model.medium
    # Set a fixed export rate
    model.reactions.EX_ac_e.lower_bound = 0.1
    model.medium = med
    assert model.reactions.EX_ac_e.lower_bound == 0.1

    # should be overwritten if actually in the medium
    med["EX_ac_e"] = 1
    model.medium = med
    assert model.reactions.EX_ac_e.lower_bound == -1


def test_medium_does_not_affect_product_exports(model: Model) -> None:
    """Test that the medium setter does not overwrite exports defined as products."""
    med = model.medium
    # Reverse reaction definition
    model.reactions.EX_ac_e *= -1
    # Set a fixed export rate
    model.reactions.EX_ac_e.bounds = -1, -0.1
    model.medium = med
    assert model.reactions.EX_ac_e.bounds == (-1, -0.1)

    # should be overwritten if actually in the medium
    med["EX_ac_e"] = 1
    model.medium = med
    assert model.reactions.EX_ac_e.upper_bound == 1



"""Provide functions for cobrapy objects to generic Python objects and vice-versa."""

from collections import OrderedDict
from operator import attrgetter, itemgetter
from typing import TYPE_CHECKING, Dict, List, Sequence, Set, Union

import numpy as np

from ..core import Gene, Metabolite, Model, Reaction
from ..util.solver import set_objective


if TYPE_CHECKING:
    from cobra import Object

_REQUIRED_REACTION_ATTRIBUTES = [
    "id",
    "name",
    "metabolites",
    "lower_bound",
    "upper_bound",
    "gene_reaction_rule",
]
_ORDERED_OPTIONAL_REACTION_KEYS = [
    "objective_coefficient",
    "subsystem",
    "notes",
    "annotation",
]
_OPTIONAL_REACTION_ATTRIBUTES = {
    "objective_coefficient": 0,
    "subsystem": "",
    "notes": {},
    "annotation": {},
}

_REQUIRED_METABOLITE_ATTRIBUTES = ["id", "name", "compartment"]
_ORDERED_OPTIONAL_METABOLITE_KEYS = [
    "charge",
    "formula",
    "_bound",
    "notes",
    "annotation",
]
_OPTIONAL_METABOLITE_ATTRIBUTES = {
    "charge": None,
    "formula": None,
    "_bound": 0,
    "notes": {},
    "annotation": {},
}

_REQUIRED_GENE_ATTRIBUTES = ["id", "name"]
_ORDERED_OPTIONAL_GENE_KEYS = ["notes", "annotation"]
_OPTIONAL_GENE_ATTRIBUTES = {
    "notes": {},
    "annotation": {},
}

_ORDERED_OPTIONAL_MODEL_KEYS = ["name", "compartments", "notes", "annotation"]
_OPTIONAL_MODEL_ATTRIBUTES = {
    "name": None,
    #  "description": None, should not actually be included
    "compartments": [],
    "notes": {},
    "annotation": {},
}


def _fix_type(
    value: Union[str, float, bool, Set, Dict],
) -> Union[str, float, bool, List, OrderedDict]:
    """Convert possible types to correct Python types.

    Parameters
    ----------
    value : str, float, bool, set, dict
        The value to fix type for.

    Returns
    -------
    str, float, bool, list, dict
        The fixed type for the value.

    """
    # Because numpy floats can not be pickled to json
    if isinstance(value, str):
        return str(value)
    if isinstance(value, float):
        return float(value)
    if isinstance(value, bool):
        return bool(value)
    if isinstance(value, set):
        return list(value)
    if isinstance(value, dict):
        return OrderedDict((key, value[key]) for key in sorted(value))
    # handle legacy Formula type
    if value.__class__.__name__ == "Formula":
        return str(value)
    if value is None:
        return ""
    return value


def _update_optional(
    cobra_object: "Object",
    new_dict: Dict,
    optional_attribute_dict: Dict,
    ordered_keys: Sequence,
) -> None:
    """Update `new_dict` with optional attributes from `cobra_object`.

    Parameters
    ----------
    cobra_object : cobra.Object
        The cobra Object to update optional attributes from.
    new_dict : dict
        The dictionary to update optional attributes for.
    optional_attribute_dict : dict
        The dictionary to use as default value lookup store.
    ordered_keys : list, tuple
        The keys to update values for.

    Raises
    ------
    IndexError
        If key in `ordered_keys` is not found in `optional_attribute_dict`.
    AttributeError
        If key in `ordered_keys` is not found in `cobra_object`.

    """
    for key in ordered_keys:
        default = optional_attribute_dict[key]
        value = getattr(cobra_object, key)
        if value is None or value == default:
            continue
        new_dict[key] = _fix_type(value)


def _metabolite_to_dict(metabolite: Metabolite) -> OrderedDict:
    """Convert a cobra Metabolite object to dictionary.

    Parameters
    ----------
    metabolite : cobra.Metabolite
        The cobra.Metabolite to convert to dictionary.

    Returns
    -------
    dict
        The converted dictionary object.

    See Also
    --------
    _metabolite_from_dict : Convert a dictionary to cobra Metabolite object.

    """
    new_metabolite = OrderedDict()
    for key in _REQUIRED_METABOLITE_ATTRIBUTES:
        new_metabolite[key] = _fix_type(getattr(metabolite, key))
    _update_optional(
        metabolite,
        new_metabolite,
        _OPTIONAL_METABOLITE_ATTRIBUTES,
        _ORDERED_OPTIONAL_METABOLITE_KEYS,
    )
    return new_metabolite


def _metabolite_from_dict(metabolite: Dict) -> Metabolite:
    """Convert a dictionary to cobra Metabolite object.

    Parameters
    ----------
    metabolite : dict
        The dictionary to convert to cobra.Metabolite .

    Returns
    -------
    cobra.Metabolite

    See Also
    --------
    _metabolite_to_dict : Convert a cobra Metabolite object to dictionary.

    """
    new_metabolite = Metabolite()
    for k, v in metabolite.items():
        setattr(new_metabolite, k, v)
    return new_metabolite


def _gene_to_dict(gene: Gene) -> OrderedDict:
    """Convert a cobra Gene object to dictionary.

    Parameters
    ----------
    gene : cobra.Gene
        The cobra.Gene to convert to dictionary.

    Returns
    -------
    dict
        The converted dictionary object.

    See Also
    --------
    _gene_from_dict : Convert a dictionary to cobra Gene object.

    """
    new_gene = OrderedDict()
    for key in _REQUIRED_GENE_ATTRIBUTES:
        new_gene[key] = _fix_type(getattr(gene, key))
    _update_optional(
        gene, new_gene, _OPTIONAL_GENE_ATTRIBUTES, _ORDERED_OPTIONAL_GENE_KEYS
    )
    return new_gene


def gene_from_dict(gene: Dict) -> Gene:
    """Convert a dictionary to cobra Gene object.

    Parameters
    ----------
    gene : dict
        The dictionary to convert to cobra.Gene .

    Returns
    -------
    cobra.Gene
        The converted cobra.Gene object.

    See Also
    --------
    _gene_to_dict : Convert a cobra Gene object to a dictionary.

    """
    new_gene = Gene(gene["id"])
    for k, v in gene.items():
        setattr(new_gene, k, v)
    return new_gene


def _reaction_to_dict(reaction: Reaction) -> OrderedDict:
    """Convert a cobra Reaction object to a dictionary.

    Parameters
    ----------
    reaction : cobra.Reaction
        The cobra.Reaction to convert to dictionary.

    Returns
    -------
    dict
        The converted dictionary object.

    See Also
    --------
    _reaction_from_dict : Convert a dictionary to a cobra Reaction object.

    """
    new_reaction = OrderedDict()
    for key in _REQUIRED_REACTION_ATTRIBUTES:
        if key != "metabolites":
            if key == "lower_bound" and (
                np.isnan(reaction.lower_bound) or np.isinf(reaction.lower_bound)
            ):
                new_reaction[key] = str(_fix_type(getattr(reaction, key)))
            elif key == "upper_bound" and (
                np.isnan(reaction.upper_bound) or np.isinf(reaction.upper_bound)
            ):
                new_reaction[key] = str(_fix_type(getattr(reaction, key)))
            else:
                new_reaction[key] = _fix_type(getattr(reaction, key))
            continue
        mets = OrderedDict()
        for met in sorted(reaction.metabolites, key=attrgetter("id")):
            mets[str(met)] = reaction.metabolites[met]
        new_reaction["metabolites"] = mets
    _update_optional(
        reaction,
        new_reaction,
        _OPTIONAL_REACTION_ATTRIBUTES,
        _ORDERED_OPTIONAL_REACTION_KEYS,
    )
    return new_reaction


def _reaction_from_dict(reaction: Dict, model: Model) -> Reaction:
    """Convert a dictionary to a cobra Reaction object.

    Parameters
    ----------
    reaction : dict
        The dictionary to convert to cobra.Reaction .
    model : cobra.Model
        The model to which the reaction should associate with.

    Returns
    -------
    cobra.Reaction
        The converted cobra.Reaction object.

    See Also
    --------
    _reaction_to_dict : Convert a cobra Reaction object to a dictionary.

    """
    new_reaction = Reaction()
    for k, v in reaction.items():
        if k in {"objective_coefficient", "reversibility", "reaction"}:
            continue
        elif k == "metabolites":
            new_reaction.add_metabolites(
                OrderedDict(
                    (model.metabolites.get_by_id(str(met)), coeff)
                    for met, coeff in v.items()
                )
            )
        else:
            if k == "lower_bound" or k == "upper_bound":
                setattr(new_reaction, k, float(v))
            else:
                setattr(new_reaction, k, v)
    return new_reaction


def model_to_dict(model: Model, sort: bool = False) -> OrderedDict:
    """Convert a cobra Model to a dictionary.

    Parameters
    ----------
    model : cobra.Model
        The model to reformulate as a dict.
    sort : bool, optional
        Whether to sort the metabolites, reactions, and genes or maintain the
        order defined in the model (default False).

    Returns
    -------
    OrderedDict
        A dictionary with keys: 'genes', 'compartments', 'id',
        'metabolites', 'notes' and 'reactions'; where 'metabolites', 'genes'
        and 'metabolites' are in turn lists of dictionaries holding all
        attributes to form the corresponding object.

    See Also
    --------
    model_from_dict : Convert a dictionary to a cobra Model.

    """
    obj = OrderedDict()
    obj["metabolites"] = list(map(_metabolite_to_dict, model.metabolites))
    obj["reactions"] = list(map(_reaction_to_dict, model.reactions))
    obj["genes"] = list(map(_gene_to_dict, model.genes))
    obj["id"] = model.id
    _update_optional(
        model, obj, _OPTIONAL_MODEL_ATTRIBUTES, _ORDERED_OPTIONAL_MODEL_KEYS
    )
    if sort:
        get_id = itemgetter("id")
        obj["metabolites"].sort(key=get_id)
        obj["reactions"].sort(key=get_id)
        obj["genes"].sort(key=get_id)
    return obj


def model_from_dict(obj: Dict) -> Model:
    """Build a cobra Model from a dictionary.

    Models stored in JSON are first formulated as a dictionary that can be read
    to a cobra Model using this function.

    Parameters
    ----------
    obj : dict
        A dictionary with keys: 'genes', 'compartments', 'id',
        'metabolites', 'notes' and 'reactions'; where 'metabolites', 'genes'
        and 'metabolites' are in turn lists of dictionaries holding all
        attributes to form the corresponding object.

    Returns
    -------
    cobra.Model
        The generated model.

    Raises
    ------
    ValueError
        If `obj` has no 'reactions' attribute.

    See Also
    --------
    model_to_dict : Convert a cobra Model to a dictionary.

    """
    if "reactions" not in obj:
        raise ValueError("Object has no .reactions attribute. Cannot load.")
    model = Model()
    model.add_metabolites(
        [_metabolite_from_dict(metabolite) for metabolite in obj["metabolites"]]
    )
    model.genes.extend([gene_from_dict(gene) for gene in obj["genes"]])
    model.add_reactions(
        [_reaction_from_dict(reaction, model) for reaction in obj["reactions"]]
    )
    objective_reactions = [
        rxn for rxn in obj["reactions"] if rxn.get("objective_coefficient", 0) != 0
    ]
    coefficients = {
        model.reactions.get_by_id(rxn["id"]): rxn["objective_coefficient"]
        for rxn in objective_reactions
    }
    set_objective(model, coefficients)
    for k, v in obj.items():
        if k in {"id", "name", "notes", "compartments", "annotation"}:
            setattr(model, k, v)
    return model



"""Test functionalities of OptGPSampler."""

import os
from typing import TYPE_CHECKING, Callable

import numpy as np
import pytest

from cobra.sampling import OptGPSampler


if TYPE_CHECKING:
    from cobra import Model
    from cobra.sampling import ACHRSampler


@pytest.fixture(scope="function")
def optgp(model: "Model") -> OptGPSampler:
    """Return OptGPSampler instance for tests."""
    sampler = OptGPSampler(model, processes=1, thinning=1)
    assert (sampler.n_warmup > 0) and (sampler.n_warmup <= 2 * len(model.variables))
    assert all(sampler.validate(sampler.warmup) == "v")

    return sampler


@pytest.mark.skipif("SKIP_MP" in os.environ, reason="unsafe for parallel execution")
def test_optgp_init_benchmark(model: "Model", benchmark: Callable) -> None:
    """Benchmark inital OptGP sampling."""
    benchmark(lambda: OptGPSampler(model, processes=2))


def test_optgp_sample_benchmark(optgp: "Model", benchmark: Callable) -> None:
    """Benchmark OptGP sampling."""
    benchmark(optgp.sample, 1)


def test_sampling(optgp: OptGPSampler) -> None:
    """Test sampling."""
    s = optgp.sample(10)
    assert all(optgp.validate(s) == "v")


def test_batch_sampling(optgp: OptGPSampler) -> None:
    """Test batch sampling."""
    for b in optgp.batch(5, 4):
        assert all(optgp.validate(b) == "v")


def test_variables_samples(achr: "ACHRSampler", optgp: OptGPSampler) -> None:
    """Test variable samples."""
    vnames = np.array([v.name for v in achr.model.variables])
    s = optgp.sample(10, fluxes=False)
    assert s.shape == (10, optgp.warmup.shape[1])
    assert (s.columns == vnames).all()
    assert (optgp.validate(s) == "v").all()


def test_reproject(optgp: OptGPSampler) -> None:
    """Test reprojection of sampling."""
    s = optgp.sample(10, fluxes=False).values
    proj = np.apply_along_axis(optgp._reproject, 1, s)
    assert all(optgp.validate(proj) == "v")

    s = np.random.rand(10, optgp.warmup.shape[1])
    proj = np.apply_along_axis(optgp._reproject, 1, s)
    assert all(optgp.validate(proj) == "v")



"""Test functionalities of I/O in YAML format."""

import json
from pathlib import Path
from typing import Callable

import pytest
from ruamel.yaml import YAML

from cobra import Model
from cobra import io as cio


def test_load_yaml_model(
    compare_models: Callable, cobra_data_directory: Path, mini_model: Model
) -> None:
    """Test the reading of YAML model."""
    yaml_model = cio.load_yaml_model(cobra_data_directory.joinpath("mini.yml"))
    assert compare_models(mini_model, yaml_model) is None


@pytest.mark.xfail(reason="schema outdated")
def test_save_yaml_model(tmp_path: Path, mini_model: Model) -> None:
    """Test the writing of YAML model."""
    jsonschema = pytest.importorskip("jsonschema")
    output_file = tmp_path.joinpath("mini.yml")
    cio.save_yaml_model(mini_model, str(output_file), sort=True)
    # validate against JSONSchema
    yaml = YAML(typ="unsafe")
    with open(output_file, "r") as infile:
        yaml_to_dict = yaml.load(infile)
    dict_to_json = json.dumps(yaml_to_dict)
    loaded = json.loads(dict_to_json)
    assert jsonschema.validate(loaded, cio.json.json_schema)
